from itertools import groupby
from decimal import Decimal
from enum import Enum
from datetime import date, datetime, time, timedelta
from pyracmon.model import Table, Column
from pyracmon.dialect.shared import MultiInsertMixin
from pyracmon.query import Q, where, holders


def read_schema(db, excludes=None, includes=None):
    """
    Collect tables in current database.

    Parameters
    ----------
    excludes: [str]
        Excluding table names.
    includes: [str]
        Including table names. If not specified, all tables are collected.

    Returns
    -------
    [Table]
        Tables.
    """
    try:
        db.stmt().execute(f"DROP TEMPORARY TABLE IF EXISTS kcu")
        db.stmt().execute(f"""\
            CREATE TEMPORARY TABLE kcu
            SELECT * FROM information_schema.key_column_usage WHERE table_schema = DATABASE()
            """)

        q = Q(excludes = excludes, includes = includes)

        cond = Q.of("c.table_schema = DATABASE()") & q.excludes.not_in("c.table_name") & q.includes.in_("c.table_name")

        w, params = where(cond)

        cursor = db.stmt().execute(f"""\
            SELECT
                c.table_name, c.column_name, c.data_type, c.column_type, c.column_key, k.referenced_table_name, k.referenced_column_name, c.extra, c.column_comment
            FROM
                information_schema.columns AS c
                LEFT JOIN kcu AS k
                    ON c.table_catalog = k.table_catalog
                        AND c.table_schema = k.table_schema
                        AND c.table_name = k.table_name
                        AND c.column_name = k.column_name
                        AND k.referenced_table_name IS NOT NULL
            {w}
            ORDER BY
                c.table_name, c.ordinal_position ASC
            """, *params)

        def map_types(t):
            base = db.context.config.type_mapping
            ptype = base and base(t)
            return ptype or _map_types(t)

        def column_of(n, t, ct, key, rt, rc, extra, comment):
            return Column(n, map_types(t), ct, key == "PRI", bool(rt), True if extra == "auto_increment" else None, comment or "")

        tables = []

        for t, cols in groupby(cursor.fetchall(), lambda row: row[0]):
            tables.append(Table(t, [column_of(*c[1:]) for c in cols]))

        cursor.close()

        if len(tables) == 0:
            return []

        cursor = db.stmt().execute(f"""\
            SELECT
                table_name, table_comment
            FROM
                information_schema.tables
            WHERE
                table_name IN ({holders(len(tables))})
            """, *[t.name for t in tables])

        table_map = {t.name: t for t in tables}

        for n, cmt in cursor.fetchall():
            if n in table_map:
                table_map[n].comment = cmt or ""

        cursor.close()

        return tables
    finally:
        db.stmt().execute(f"DROP TEMPORARY TABLE IF EXISTS kcu")


def _map_types(t):
    if t == "tinyint" or t == "smallint" or t == "mediumint" or t == "int" or t == "bigint":
        return int
    elif t == "decimal":
        return Decimal
    elif t == "float" or t == "double":
        return float
    elif t == "bit":
        return int
    elif t == "char" or t == "varchar" or t == "binary" or t == "varbinary" or t == "text":
        return str
    elif t == "blob":
        return bytes
    elif t == "enum":
        return Enum
    elif t == "date":
        return date
    elif t == "datetime" or t == "timestamp":
        return datetime
    else:
        return object


class MySQLMixin(MultiInsertMixin):
    @classmethod
    def last_sequences(cls, db, num):
        cols = [c for c in cls.columns if c.incremental]

        if len(cols) > 1:
            raise ValueError(f"MySQL allows tables having only an auto-increment column.")
        elif len(cols) == 1:
            d = db.cursor()
            d.execute(f"SELECT LAST_INSERT_ID()")
            sequence = d.fetchone()[0] + num - 1
            d.close()
            return [(cols[0], sequence)]
        else:
            return []


mixins = [MySQLMixin]


def found_rows(db):
    with db.cursor() as c:
        c.execute("SELECT FOUND_ROWS()")
        return c.fetchone()[0]


from functools import reduce
from .sql import Sql
from .marker import Marker


class Q:
    """
    The instance of this class holds parameters to build query conditions.

    Each parameter passed by the constructor becomes an instance method of created instance,
    which takes a condition clause including placeholders which will accept the parameter in query execution.
    `$_` in the clause will be replaced with correct placeholder marker in query execution by `Statement`.

    >>> q = Q(a=1)
    >>> q.a("a = $_")
    Condition: 'a = $_' -- [1]

    Method whose name is not passed by the constructor returns empty condition which has no effect on the query.

    >>> q.b("b = $_")
    Condition: '' -- []

    This feature simplifies a query construction in case some parameters can be absent.

    >>> def search(db, q):
    >>>     w, params = where(q.a("a = $_") & q.b("b = $_"))
    >>>     db.stmt().execute(f"SELECT * FROM table {w}", *params)
    >>> 
    >>> search(db, Q(a=1))        # SELECT * FROM table WHERE a = 1
    >>> search(db, Q(a=1, b=2)) # SELECT * FROM table WHERE a = 1 AND b = 2
    >>> search(db, Q())             # SELECT * FROM table

    Additionally, this class provides utility class methods creating conditions.

    Using `of()` is the most simple way to create a condition clause.

    >>> Q.of("a = $_", 1)
    Condition: 'a = $_' -- [1]

    Other utility methods correspond to basic operators defined in SQL.
    They takes keyword arguments and create conditions by applying operator to each item respectively.

    >>> Q.eq(a=1)
    Condition: 'a = %s' -- [1]
    >>> Q.eq(a=1, b=2)
    Condition: 'a = %s AND b = %s' -- [1, 2]
    >>> Q.in_(a=[1, 2, 3])
    Condition: 'a IN (%s, %s, %s)' -- [1, 2, 3]
    >>> Q.like(a="abc")
    Condition: 'a LIKE %s' -- ["%abc%"]
    """
    class Attribute:
        def __init__(self, value):
            self.value = value

        def __call__(self, expression, convert=None):
            """
            Craetes conditional object composed of given clause and the attribute value as parameters.

            Parameters
            ----------
            clause: str | object -> str
                A clause or a function generating a clause by taking the attribute value.
            convert: object -> object
                A function converting the attribute value to parameters.
                If this function returns a value which is not a list, a list having only the value is used.

            Returns
            -------
            Conditional
                Condition.
            """
            expression = expression if isinstance(expression, str) else expression(self.value)
            params = convert(self.value) if convert else [self.value]

            return Conditional(expression, params if isinstance(params, list) else [params])

        @property
        def all(self):
            return Q.CompositeAttribute(self.value, True)

        @property
        def any(self):
            return Q.CompositeAttribute(self.value, False)

        def __getattr__(self, key):
            method = getattr(Q, key)
            def invoke(col, convert=None):
                return method(**{col: convert(self.value) if convert else self.value})
            return invoke

    class CompositeAttribute(Attribute):
        def __init__(self, value, and_):
            super().__init__(value)
            self._and = and_

        def __call__(self, expression, convert=None):
            conds = [Q.Attribute(v)(expression, convert) for v in self.value]
            return Conditional.all(conds) if self._and else Conditional.any(conds)

        def __getattr__(self, key):
            method = getattr(Q, key)
            def invoke(col, convert=None):
                conds = [method(**{col: convert(v) if convert else v}) for v in self.value]
                return Conditional.all(conds) if self._and else Conditional.any(conds)
            return invoke

    class NoAttribute:
        def __call__(self, expression, holder=lambda x:x):
            return Conditional()

        @property
        def all(self):
            return self

        @property
        def any(self):
            return self

        def __getattr__(self, key):
            method = getattr(Q, key)
            def invoke(col, convert=None):
                return Conditional()
            return invoke

    def __init__(self, __include_none=False, **kwargs):
        self.attributes = dict([(k, v) for k, v in kwargs.items() if __include_none or v is not None])

    def __getattr__(self, key):
        if key in self.attributes:
            return Q.Attribute(self.attributes[key])
        else:
            return Q.NoAttribute()

    @classmethod
    def of(cls, expression="", params=[]):
        """
        Creates a condition directly from an expression and parameters.

        Parameters
        ----------
        expression: str
            Condition expression.
        params: [object]
            Parameters used in the condition.

        Returns
        -------
        Conditional
            Condition.
        """
        if isinstance(params, list):
            pass
        elif isinstance(params, tuple):
            params = list(params)
        else:
            params = [params]
        return Conditional(expression, params)

    @classmethod
    def eq(cls, __alias=None, __and=True, **kwargs):
        """
        Creates a condition applying '=' operator to columns.

        Parameters
        ----------
        __alias: str
            Alias prepended to the column.
        __and: bool
            If `True`, conditions are concatenated by `AND`, otherwise `OR`.
        kwargs: {str:object}
            Mapping from columns to values.

        Returns
        -------
        Conditional
            Condition.
        """
        def is_null(col, val):
            if val is None:
                return f"{col} IS NULL", []
            elif val is True:
                return f"{col}", []
            elif val is False:
                return f"NOT {col}", []
            return None
        return _conditional("=", __and, kwargs, is_null, __alias)

    @classmethod
    def neq(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies `!=`.
        """
        def is_null(col, val):
            if val is None:
                return f"{col} IS NOT NULL", []
            elif val is True:
                return f"NOT {col}", []
            elif val is False:
                return f"{col}", []
            return None
        return _conditional("!=", __and, kwargs, is_null, __alias)

    @classmethod
    def in_(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies `IN`.

        If parameters are empty, this method returns a condition being never satisfied.
        """
        def in_list(col, val):
            if len(val) == 0:
                return "1 = 0", []
            else:
                holder = ', '.join(['$_'] * len(val))
                return f"{col} IN ({holder})", val
        return _conditional("IN", __and, kwargs, in_list, __alias)

    @classmethod
    def not_in(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies `NOT IN`.

        If parameters are empty, this method returns a condition being always satisfied.
        """
        def in_list(col, val):
            if len(val) == 0:
                return "", []
            else:
                holder = ', '.join(['$_'] * len(val))
                return f"{col} NOT IN ({holder})", val
        return _conditional("NOT IN", __and, kwargs, in_list, __alias)

    @classmethod
    def match(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies 'LIKE'. Parameter is passed to query without being modified.
        """
        return _conditional("LIKE", __and, kwargs, None, __alias)

    @classmethod
    def like(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies 'LIKE'. Parameter is escaped and enclosed with wildcards (%) to execute partial match.
        """
        return _conditional("LIKE", __and, {k: f"%{escape_like(v)}%" for k, v in kwargs.items()}, None, __alias)

    @classmethod
    def startswith(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies 'LIKE'. Parameter is escaped and appended with wildcards (%) to execute prefix match.
        """
        return _conditional("LIKE", __and, {k: f"{escape_like(v)}%" for k, v in kwargs.items()}, None, __alias)

    @classmethod
    def endswith(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies 'LIKE'. Parameter is escaped and prepended with wildcards (%) to execute backward match.
        """
        return _conditional("LIKE", __and, {k: f"%{escape_like(v)}" for k, v in kwargs.items()}, None, __alias)

    @classmethod
    def lt(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies '<'.
        """
        return _conditional("<", __and, kwargs, None, __alias)

    @classmethod
    def le(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies '<='.
        """
        return _conditional("<=", __and, kwargs, None, __alias)

    @classmethod
    def gt(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies '>'.
        """
        return _conditional(">", __and, kwargs, None, __alias)

    @classmethod
    def ge(cls, __alias=None, __and=True, **kwargs):
        """
        Works like `eq`, but applies '>='.
        """
        return _conditional(">=", __and, kwargs, None, __alias)


def _conditional(op, and_, column_values, gen=None, alias=None):
    cond = Conditional()

    def concat(c):
        nonlocal cond
        if and_:
            cond &= c
        else:
            cond |= c

    for col, val in column_values.items():
        col = f"{alias}.{col}" if alias else col

        if gen:
            r = gen(col, val)
            if r is not None:
                concat(Conditional(r[0], r[1]))
                continue

        concat(Conditional(f"{col} {op} $_", [val]))

    return cond


class Expression:
    def __init__(self, expression, params):
        self.expression = expression
        self.params = params


class Conditional(Expression):
    """
    Represents a single condition composed of an expression and parameters used for placeholders in the expression.

    Parameters must be a list where the index of each parameter matches the index of placeholder for it.
    The expression accepts only the automatic numbering template parameter `$_`.

    Logical operators such as `&`, `|` and `~` is available to generate new condition.

    >>> c1 = Q.of("a = $_", 0)
    >>> c2 = Q.of("b < $_", 1)
    >>> c3 = Q.of("c > $_", 2)
    >>> c = ~(c1 & c2 | c3)
    >>> c
    Condition: NOT (((a = $_) AND (b < $_)) OR (c > $_)) -- [0, 1, 2]
    """
    @classmethod
    def all(cls, conditionals):
        """
        Concatenates conditional objects with `AND`.

        Parameters
        ----------
        conditionals: [Conditional]
            Condition.

        Returns
        -------
        Conditional
            Concatenated conditional object
        """
        return reduce(lambda acc, c: acc & c, conditionals, Conditional())

    @classmethod
    def any(cls, conditionals):
        """
        Concatenates conditional objects with `OR`.

        Parameters
        ----------
        conditionals: [Conditional]
            Condition.

        Returns
        -------
        Conditional
            Concatenated conditional object
        """
        return reduce(lambda acc, c: acc | c, conditionals, Conditional())

    def __init__(self, expression="", params=None):
        super().__init__(expression, params or [])

    def __repr__(self):
        return f"Condition: '{self.expression}' -- {self.params}"

    def __call__(self, marker):
        """
        Deprecated.
        """
        c, p = Sql(marker, self.expression).render(*self.params)
        if not isinstance(p, list):
            raise ValueError(f"Only list style marker is available.")
        return Conditional(c, p)

    def __and__(self, other):
        expression = ""
        if self.expression and other.expression:
            expression = f"({self.expression}) AND ({other.expression})"
        elif self.expression:
            expression = self.expression
        elif other.expression:
            expression = other.expression

        return Conditional(expression, self.params + other.params)

    def __or__(self, other):
        expression = ""
        if self.expression and other.expression:
            expression = f"({self.expression}) OR ({other.expression})"
        elif self.expression:
            expression = self.expression
        elif other.expression:
            expression = other.expression

        return Conditional(expression, self.params + other.params)

    def __invert__(self):
        if self.expression:
            return Conditional(f"NOT ({self.expression})", self.params)
        else:
            return Conditional(f"1 = 0", [])


def escape_like(v):
    """
    Escape characters for the use in `LIKE` condition.

    Parameters
    ----------
    v: str
        A string.

    Returns
    -------
    str
        Escaped string.
    """
    def esc(c):
        if c == "\\":
            return r"\\\\"
        elif c == "%":
            return r"\%"
        elif c == "_":
            return r"\_"
        else:
            return c
    return ''.join(map(esc, v))


def where(condition):
    """
    Generates a where clause representing given condition.

    Parameters
    ----------
    condition: Conditional
        Condition.

    Returns
    -------
    str
        A where clause starting from `WHERE` or empty string if the condition is empty.
    [object]
        Parameters for placeholders in where clause.
    """
    return ('', []) if condition.expression == '' else (f'WHERE {condition.expression}', condition.params)


def order_by(columns):
    """
    Generates ORDER BY clause from columns and directions.

    Parameters
    ----------
    columns: { str: bool }
        An ordered dictionary mapping column name to its direction, where `True` denotes `ASC` and `False` denotes `DESC`.

    Returns
    -------
    str
        ORDER BY clause.
    """
    def col(cd):
        return f"{cd[0]} ASC" if cd[1] else f"{cd[0]} DESC"
    return '' if len(columns) == 0 else f"ORDER BY {', '.join(map(col, columns.items()))}"


def ranged_by(limit=None, offset=None):
    """
    Generates LIMIT and OFFSET clause using marker.

    Parameters
    ----------
    limit: int
        Limit value or `None`. 
    offset: int
        OFfset value or `None`.

    Returns
    -------
    str
        LIMIT and OFFSET clause.
    [object]
        Parameters for placeholders in LIMIT and OFFSET clause.
    """
    clause, params = [], []

    if limit is not None:
        clause.append("LIMIT $_")
        params.append(limit)

    if offset is not None:
        clause.append("OFFSET $_")
        params.append(offset)

    return ' '.join(clause) if clause else '', params


def holders(length_or_keys, qualifier=None):
    """
    Generates partial query string containing placeholders separated by comma.

    Parameters
    ----------
    length_or_keys: int | [str]
        The number of placeholders or keys assigned to them.
    qualifier: {int: str -> str}
        Mapping from indexes to functions. Each function converts placeholder string at paired index.

    Returns
    -------
    str
        Query string.
    """
    if isinstance(length_or_keys, int):
        hs = ["${_}"] * length_or_keys
    else:
        def key(k):
            if isinstance(k, int):
                return f"${{_{k}}}"
            elif k:
                return f"${{{k}}}"
            else:
                return "${_}"
        hs = [key(k) for k in length_or_keys]

    if qualifier:
        hs = [qualifier.get(i, _noop)(h) for i, h in enumerate(hs)]

    return ', '.join(hs)


def values(length_or_key_gen, rows, qualifier=None):
    """
    Generates partial query string for `VALUES` clause in insertion query.

    Parameters
    ----------
    length_or_key_gen: int | [int -> str]
        The number of placeholders or functions which takes a row index and returns a key.
    rows: int
        The number of rows to insert.
    qualifier: {int: str -> str}
        Mapping from indexes to functions. Each function converts placeholder string at paired index.

    Returns
    -------
    str
        Query string.
    """
    if isinstance(length_or_key_gen, int):
        lok = (lambda i: length_or_key_gen)
    else:
        lok = (lambda i: [g(i) for g in length_or_key_gen])

    return ', '.join([f"({holders(lok(i), qualifier)})" for i in range(rows)])


class QueryHelper:
    """
    Deprecated.
    """
    def __init__(self, api, config):
        self.api = api
        self.config = config

    def marker(self):
        """
        Deprecated. Don't create marker.
        """
        # TODO read paramstyle from config.
        return Marker.of(self.api.paramstyle)

    def holders(self, keys, qualifier = None, start = 0, marker = None):
        """
        Deprecated. Use global `holders` instead.
        """
        m = marker or self.marker()
        qualifier = qualifier or {}

        if isinstance(keys, int):
            return ', '.join([qualifier.get(i, _noop)(m(i + start)) for i in range(keys)])
        else:
            return ', '.join([qualifier.get(i, _noop)(m(k)) for i, k in enumerate(keys)])

    def values(self, keys, rows, qualifier = None, start = 0, marker = None):
        """
        Deprecated. Use global `values` instead.
        """
        num = keys if isinstance(keys, int) else len(keys)
        m = marker or self.marker()
        return ', '.join([f"({self.holders(keys, qualifier, start + num * i, m)})" for i in range(0, rows)])


def _noop(x):
    return x



from inspect import signature, Signature
from typing import TypeVar, get_type_hints
from .template import GraphTemplate
from .graph import Node
from .schema import Shrink, Extend, T, TypedDict, Typeable, GraphSchema
from .util import chain_serializers, wrap_serializer


class S:
    """
    An utility class to build `NodeSerializer`.
    """
    @classmethod
    def of(cls, namer=None, aggregator=None, *serializers):
        """
        Create an instance of `NodeSerializer`.

        Parameters
        ----------
        namer: str | str -> str
            A string or naming function.
        aggregator: [T] -> T | int
            An aggregation function or an index at which the value is selected from converted values in aggregation phase.
        serializer: T -> U
            A list of *serializer*s.

        Returns
        -------
        NodeSerializer
            Created instance.
        """
        return NodeSerializer(namer, aggregator, *serializers)

    @classmethod
    def builder(cls, builder):
        """
        This method is used as decorator to put decorating target to `S` builder methods.

        >>> @S.builder
        >>> def some_func():
        >>>     ...
        >>>
        >>> S.some_func()

        Parameters
        ----------
        builder: callable
            Decorating target.

        Returns
        -------
        callable
            Decorated builder.
        """
        def f(*args, **kwargs):
            return builder(S.of(), *args, **kwargs)
        setattr(S, builder.__name__, f)
        return builder


class NodeSerializer:
    """
    The instance of this class contains the configurations to convert a `NodeContainer` into a serializable value.

    The conversion is composed of 3 phases.

    1. If `aggregator` is set, aggregate the list of nodes into a node or shrinked list of nodes.
    2. Convert the entity of each node into a serializable value by *serializer*.
    3. Put the converted value(s) into the dictionary converted from the parent node with a key determined by `namer`.

    These phases are applied from the roots of graph to their descendants as long as the entity is converted to `dict`.
    Children of the entity which is not converted to `dict` are simply ignored.

    When `namer` is a callable, the result updates parent dictionary with its items instead of just being put to it.

    >>> graph.append(a=dict(a1=1, a2=2), b=dict(b1=3, b2=4))
    >>> GraphSpec().to_dict(
    >>>     graph.view,
    >>>     a = S.of(),
    >>>     b = S.merge(lambda n: f"__{n}__"),
    >>> )
    {"a": [{"a1":1, "a2":2, "__b1__":3, "__b2__":4}]}

    In this case, the property name `b` is ignored and `head()` aggregation is implicitly applied to the property if no aggregator is set.
    As shown in the example, the converted value on `b` must be a `dict` in order to being expanded to key-value pairs.
    Exception raises when the value of another type is obtained.
    """
    def __init__(self, namer=None, aggregator=None, *serializers):
        self._namer = namer
        self._aggregator = aggregator
        self._serializers = list(serializers)
        self._doc = ""

    @property
    def namer(self):
        """
        Naming function.

        Returns
        -------
        str -> str
            A function which determines the key of the property even when `namer` is set to a string.
        """
        if not self._namer:
            return lambda n: n
        elif isinstance(self._namer, str):
            return lambda n: self._namer
        else:
            return lambda n: self._namer(n)

    @property
    def aggregator(self):
        """
        Aggregation function.

        Returns
        -------
        [Node] -> Node | [Node]
            A function to aggregate a list of nodes to a node or another list of nodes.
        """
        if self._aggregator is None:
            def agg(values: [T]) -> [T]:
                return values
            return agg
        elif signature(self._aggregator).return_annotation == Signature.empty:
            def agg(values: [T]) -> [T]:
                return self._aggregator(values)
            return agg
        else:
            return self._aggregator

    @property
    def serializer(self):
        """
        Serialization function.

        Returns
        -------
        (SerializationContext, Node, object -> object, object) -> object
            A function to convert an entity value into a serializable value.
        """
        return chain_serializers(self._serializers)

    @property
    def be_merged(self):
        """
        Returns whether the converted value will be merged into parent.

        Returns
        -------
        bool
            `True` if converted value will be merged into parent.
        """
        return callable(self._namer)

    @property
    def be_singular(self):
        """
        Returns whether the converted value will be a singular object, not a list.

        This is estimated by annotation of aggregation function. If its returning type is not annotated, this property always returns `False`.
        `S` builder methods adds appropriate annotation to given function when it does not have the annotation.

        Returns
        -------
        bool
            `True` if converted value will be a singular object. `False` means that it will be a list.
        """
        return not isinstance(signature(self.aggregator).return_annotation, list)

    def _set_aggregator(self, aggregator, folds):
        try:
            rt = signature(aggregator).return_annotation
        except:
            rt = Signature.empty

        if rt == Signature.empty:
            def agg(vs: [T]) -> (T if folds else [T]):
                return aggregator(vs)
            self._aggregator = agg
        elif isinstance(rt, list) ^ (not folds):
            raise ValueError(f"Return annotation of function is not valid.")
        else:
            self._aggregator = aggregator

        return self

    #----------------------------------------------------------------
    # Documentation
    #----------------------------------------------------------------
    @S.builder
    def doc(self, document):
        """
        Set the documentation for this node.

        Parameters
        ----------
        document: str
            A documentation string.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        self._doc = document
        return self

    #----------------------------------------------------------------
    # Naming
    #----------------------------------------------------------------
    @S.builder
    def name(self, name):
        """
        Set a key in parent dictionary.

        Parameters
        ----------
        name: str
            A key string.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        if not isinstance(name, str):
            raise ValueError(f"The name of node must be a string but {type(name)} is given.")
        self._namer = name
        return self

    @S.builder
    def merge(self, namer=None):
        """
        Set a naming function taking a property name and returning a key in parent `dict`.

        `NodeSerializer` built with this methods merges the converted `dict` into parent `dict` and its `be_merged` property becomes `True`.

        Because merging needs folding, this method overrides the aggregation function by invoking `head()` internally
        if this serializer is not configured to fold nodes into a single node.

        Parameters
        ----------
        namer: str -> str
            The naming function. If `None`, the property name is returned as it is.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        if namer and not callable(namer):
            raise ValueError(f"The method merging a node into its parent node must be callable or None.")
        self._namer = namer or (lambda x:x)
        if not self.be_singular:
            self.head()
        return self

    #----------------------------------------------------------------
    # Aggregation
    #----------------------------------------------------------------
    @S.builder
    def at(self, index, alt=None):
        """
        Set an aggregator which picks up the node at the index.

        Parameters
        ----------
        index: int
            An index of the element.
        alt: object
            A value used when no node is found at the index.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        return self.fold(lambda vs: vs[index] if len(vs) > index else alt)

    @S.builder
    def head(self, alt=None):
        """
        Set an aggregator which picks up the first node.

        Parameters
        ----------
        alt: object
            A value used when there exists no nodes.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        return self.at(0, alt)

    @S.builder
    def last(self, alt=None):
        """
        Set an aggregator which picks up the last node.

        Parameters
        ----------
        alt: object
            A value used when there exists no nodes.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        return self.fold(lambda vs: vs[-1] if len(vs) > 0 else alt)

    @S.builder
    def fold(self, aggregator):
        """
        Set an aggregation function converting a list of nodes into a single node.

        Parameters
        ----------
        aggregator: [Node] -> Node
            An aggregation function.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        return self._set_aggregator(aggregator, True)

    @S.builder
    def select(self, aggregator):
        """
        Set an aggregation function chooding a list of nodes from all nodes in `NodeContainer`.

        Parameters
        ----------
        aggregator: [Node] -> [Node]
            An aggregation function.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        return self._set_aggregator(aggregator, False)

    #----------------------------------------------------------------
    # Serizlization
    #----------------------------------------------------------------
    @S.builder
    def each(self, func):
        """
        Set a function converting a node entity into a serializable value.

        In order to progress serialization to child nodes, the function MUST returns a `dict`.

        The function will be invoked with 0 to 4 arguments listed below.

        - `SerializationContext` of the serialization.
        - `Node` to serialize.
        - A function which takes the entity value and returns converted value by applying all serialization functions added beforehand.
        - An entity value to convert.

        When the number of arguments in signature of the function is less than 4, former arguments is the list are omitted.
        For example, `def func(c, n, b, v):` is invoked with all arguments, while `def func(b, v):` is invoked with only 3rd and 4th arguments.

        Parameters
        ----------
        func: (SerializationContext, Node, object -> object, object) -> object
            A function converting a node entity into a serializable value.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        self._serializers.append(func)
        return self

    @S.builder
    def sub(self, **settings):
        """
        Set serialization settings to serializer sub graph.

        This method is used for the property whose kind is `GraphTemplate`.
        `settings` keyword arguments should be the same form of the keyword arguments of `to_dict()` of `GraphSpec`.

        Parameters
        ----------
        settings: {str: NodeSerializer}
            Serialization settings used to serialize sub graph.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        class SubGraph(Typeable[T]):
            serializers = settings.copy()

            @staticmethod
            def resolve(sub_graph, bound, arg, spec):
                return GraphSchema(spec, arg.template, **sub_graph.serializers).schema

        def to_dict(c, n, b, v) -> SubGraph[T]:
            vv = b(v)
            return SerializationContext(settings, c.finder).execute(vv.view)
        self._serializers.append(to_dict)
        return self.head()

    @S.builder
    def alter(self, generator=None, excludes=None, includes=None):
        """
        Extends and shrinks the dictionary.

        Parameters
        ----------
        generator: (SerializationContext, Node, object -> object, object) -> {str: object}
            A function generating dictionary
        excludes: Iterable[str]
            Keys to exclude from the dictionary.
        includes: Iterable[str]
            Keys to keep in the dictionary.

        Returns
        -------
        NodeSerializer
            This instance.
        """
        excludes = excludes or []

        class EachExtend(Extend[T]):
            @classmethod
            def schema(cls, bound, arg):
                return signature(generator).return_annotation if generator else Signature.empty

        class EachShrink(Shrink[T]):
            @classmethod
            def select(cls, td, bound):
                return excludes, includes

        def convert(c, n, b, v) -> EachShrink[EachExtend[T]]:
            ext = wrap_serializer(generator)(c, n, b, v) if generator else {}
            vv = b(v)
            vv.update(**ext)
            return {k:v for k, v in vv.items() if (not includes or k in includes) and k not in excludes}

        self._serializers.append(convert)
        return self


class SerializationContext:
    """
    This class provides a functionality to serialize a graph by using containing `NodeSerializer` s.
    """
    def __init__(self, settings, finder, node_params=None):
        self.serializer_map = {n:self._to_serializer(s) for n, s in settings.items()}
        self.finder = finder
        self.node_params = node_params or {}

    def __getitem__(self, node):
        """
        Returns an accessor to parameters for given node.

        Parameters
        ----------
        node: Node | str
            Node or node name.

        Returns
        -------
        Accessor
            An object exposing parameters via its attributes of their names.
        """
        name = node.name if isinstance(node, Node) else node
        params = self.node_params.get(name, {})

        class Accessor:
            def __getattr__(self, key):
                return params.get(key, None)

        return Accessor()

    def _to_serializer(self, s):
        if isinstance(s, NodeSerializer):
            return s
        else:
            return S.of(*s)

    def execute(self, graph):
        """
        Serializes a graph.

        Parameters
        ----------
        graph: GraphView
            The view of graph to serialize.

        Returns
        -------
        dict
            Serialization result.
        """
        result = {}

        for n, c in filter(lambda nc: nc[1]().property.parent is None, graph):
            self.serialize_to(c().name, c, result)

        return result

    def serialize_to(self, name, container, parent):
        """
        Convert nodes into serializable values and append them into the parent dictionary.

        Parameters
        ----------
        name: str
            A property name for the nodes.
        container: NodeContainer | Node.Children
            Container of nodes. 
        parent: dict
            A parent dictionary the converted values are appended to.
        """
        ns = self.serializer_map.get(name, None)

        if ns:
            agg = ns.aggregator(container().nodes)

            serializer = ns.serializer

            if ns.be_singular:
                # Alternative value given to aggregation function may be returned instead of node.
                value = self._serialize_node(container().property, agg, serializer) if isinstance(agg, Node) else agg

                if ns.be_merged:
                    if not isinstance(value, dict):
                        raise ValueError(f"Serialized value must be dict but {type(value)}.")

                    parent.update({ns.namer(k):v for k, v in value.items()})
                else:
                    parent[ns.namer(name)] = value
            else:
                if ns.be_merged:
                    raise ValueError(f"Merging to parent dict requires folding.")

                parent[ns.namer(name)] = [self._serialize_node(container().property, n, serializer) for n in agg]

    def _serialize_node(self, prop, node, serializer):
        value = serializer(self, node, chain_serializers(self.finder(prop.kind)), node.entity)

        if isinstance(value, dict):
            for n, ch in node.children.items():
                self.serialize_to(n, ch.view, value)

        return value


from string import digits, Template
from .marker import Marker


class Sql:
    """
    Provides functionalities to render SQL string from the template containing place holders.

    SQL rendering is conform to the way of `string.Template` which replaces `$` prefixed variables with parameters.
    """
    class Substitute:
        def __init__(self, marker):
            self.marker = marker

        def __getitem__(self, key):
            if key == "_":
                return self.marker()
            elif key[0] == "_" and all([c in digits for c in key[1:]]):
                return self.marker(int(key[1:]))
            else:
                return self.marker(key)

    def __init__(self, marker, template):
        self.marker = marker
        self.template = template

    def render(self, *args, **kwargs):
        """
        Renders SQL and converts parameters into the form available for current database driver.

        Parameters
        ----------
        params: object
            Parameters. Raising an exception when the type of this object does not conform to the paramstyle.

        Returns
        -------
        str
            SQL string.
        object
            Values available as parameters of the SQL.
        """
        self.marker.reset()

        sub = Sql.Substitute(self.marker)

        return Template(self.template).substitute(sub), self.marker.params(*args, **kwargs)





from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from uuid import UUID, uuid1, uuid3
from .util import Matcher, config


class TestingState:
    indexes = {}

    @classmethod
    def reset(cls, model=None):
        cls.indexes.clear()

    @classmethod
    def inc(cls, model, count):
        v = cls.indexes.setdefault(model, 1)
        cls.indexes[model] = v + count
        return v

    @classmethod
    def set_index(cls, model, index):
        cls.indexes[model] = index


class TestingMixin:
    """
    Mixin class for model types providing methods designed for testing.
    """
    @classmethod
    def by(cls, index):
        """
        Set current fixture index.

        Parameters
        ----------
        index: int
            Fixture index.

        Returns
        -------
        type[TestingMixin]
            This type.
        """
        TestingState.set_index(cls, index)
        return cls

    @classmethod
    def fixture(cls, db, variable=None, index=None):
        """
        Insert record with auto-generated column values.

        Parameters
        ----------
        db: Connection
            DB connection. If a value evaluated to be `False` in boolean context, generated model is not inserted and just returned.
        variable: int | dict | model | [dict] | [model]
            When `int`, inserts records as many as the number. All of their column values are generated.
            When `dict`, model object or a list of them, inserts record(s) represented by them. Unspecified column values are generated.
        index: int
            Use this to specify index used to generate column values explicitly. If set, indexing state is not updated.

        Returns
        -------
        model | [model]
            Inserted model(s).
        """
        if variable is None or isinstance(variable, int):
            num = variable or 1
            index = TestingState.inc(cls, num) if index is None else index
            models = [_generate_model(cls, index+i) for i in range(num)]
            if db:
                cls.inserts(db, models)
            return models
        elif isinstance(variable, (cls, dict)):
            num = 1
            index = TestingState.inc(cls, num) if index is None else index
            model = _generate_model(cls, index, variable)
            if db:
                cls.insert(db, _generate_model(cls, index, variable))
            return model
        elif isinstance(variable, list):
            num = len(variable)
            index = TestingState.inc(cls, num) if index is None else index
            models = [_generate_model(cls, index+i, v) for i, v in enumerate(variable)]
            if db:
                cls.inserts(db, models)
            return models
        else:
            raise ValueError(f"Second argument of fixture() must be an int, dict, model or list of dict or model but {type(variable)} is passed.")

    def match(self, **expected):
        """
        Tests columns values matches to expected values.

        Parameters
        ----------
        expected: {str: (object | Matcher)}
            Expected values.

        Returns
        -------
        bool
            Matches or not.
        """
        for k, v in expected.items():
            actual = getattr(self, k)

            if isinstance(v, Matcher):
                if not v.match(actual):
                    return False
            else:
                if v != actual:
                    return False
        return True


def _generate_model(model, index, model_or_dict=None):
    values = {}

    if isinstance(model_or_dict, TestingMixin):
        values = {c.name:v for c, v in model_or_dict}
    elif isinstance(model_or_dict, dict):
        values = dict(**model_or_dict)

    values.update(**{c.name:_generate_value(model.table, c, index) for c in model.columns if c.name not in values and not c.pk})

    return model(**values)


fixed_uuid=uuid1(0, 0)


def _generate_value(table, column, index):
    """
    Generates a value for the column on an index.

    Parameters
    ----------
    table: Table
        Table.
    column: Column
        Column.
    index: int
        Index

    Returns
    -------
    object
        Generated value.
    """
    mapping = config().fixture_mapping
    if mapping:
        value = mapping(table, column, index)
        if value is not None:
            return value

    tz_aware = config().fixture_tz_aware

    if column.ptype is bool:
        return True
    elif column.ptype is float:
        return float(f"{index}.{index+1}")
    elif column.ptype is int:
        return index
    elif column.ptype is Decimal:
        return Decimal(index)
    elif column.ptype is str:
        return f"{column.name}-{index}"
    elif column.ptype is bytes:
        return f"{column.name}-{index}".encode()
    elif column.ptype is date:
        return date.today()
    elif column.ptype is datetime:
        return datetime.now().astimezone() if tz_aware else datetime.now()
    elif column.ptype is time:
        return datetime.now().astimezone().time() if tz_aware else datetime.now().time()
    elif column.ptype is timedelta:
        return timedelta(days=index+1)
    elif column.ptype is UUID:
        return str(uuid3(fixed_uuid, f"{table.name}-{column.name}-{index}"))
    elif isinstance(column.ptype, type) and issubclass(column.ptype, Enum):
        return next(iter(column.ptype), None)
    else:
        return None



from functools import reduce
from .model import *
from .select import *
from .query import *
from .util import key_to_index


class CRUDMixin(SelectMixin):
    """
    This class provides class methods available on all model types.

    Every method takes the DB connection object as its first argument.

    Also, arguments listed below are commonly used in some methods of this mixin class:

    - `pks` represents value(s) of primary key(s).
        - If `pks` is a dictionary, each item is considered to be a pair of column name and value respectively.
        - Otherwise, the model must have a single primary key and `pks` is its value.
    - `qualifier` is a dictionary whose key is a column name and the value is a function generating SQL expression around the placeholder for the column.
        - This argument is used to overwrite the expression generated by `Marker` in insert and update query.
        - The function must take an argument to which the default expression is passed.
        - Be aware that the actual value stored in database can be different from value set on model when this argument exists.
    """
    @classmethod
    def count(cls, db, condition=Q.of()):
        """
        Count rows which satisfies the condition.

        Parameters
        ----------
        db: Connection
            DB connection.
        condition: Conditional
            Condition.

        Returns
        -------
        int
            The number of rows.
        """
        wc, wp = where(condition)
        c = db.stmt().execute(f"SELECT COUNT(*) FROM {cls.name}{_spacer(wc)}", *wp)
        return c.fetchone()[0]

    @classmethod
    def fetch(cls, db, pks, lock=None):
        """
        Fetch a record by primary key(s).

        Parameters
        ----------
        db: Connection
            DB connection.
        pks: object | {str: object}
            Representation of primary key(s).
        lock: object
            An object whose string expression is appended to query.

        Returns
        -------
        cls
            A model of the record.
        """
        cols, vals = parse_pks(cls, pks)
        cond = Conditional.all([Q.eq(**{c: v}) for c, v in zip(cols, vals)])
        wc, wp = where(cond)
        s = cls.select()
        c = db.stmt().execute(f"SELECT {s} FROM {cls.name}{_spacer(wc)}{_spacer(lock)}", *wp)
        row = c.fetchone()
        return read_row(row, s)[0] if row else None

    @classmethod
    def fetch_where(cls, db, condition=Q.of(), orders={}, limit=None, offset=None, lock=None):
        """
        Fetch records which satisfy a condition.

        Parameters
        ----------
        db: Connection
            DB connection.
        condition: Conditional
            Condition.
        orders: {str: bool}
            Ordering parameters. Each key is column and its value denotes direction; `True` is ascending and `False` is descending.
        limit: int
            The number of rows to fetch. If `None`, all rows are obtained.
        offset: int
            The number of rows to skip.
        lock: object
            An object whose string expression is appended to query.

        Returns
        -------
        [cls]
            Models of records.
        """
        wc, wp = where(condition)
        rc, rp = ranged_by(limit, offset)
        s = cls.select()
        c = db.stmt().execute(f"SELECT {s} FROM {cls.name}{_spacer(wc)}{_spacer(order_by(orders))}{_spacer(rc)}{_spacer(lock)}", *(wp + rp))
        return [read_row(row, s)[0] for row in c.fetchall()]

    @classmethod
    def insert(cls, db, record, qualifier={}):
        """
        Insert a record.

        Returned model object contains auto incremental column even if they are not set beforehand.
        On the contrary, default value defined on column is not set.

        Parameters
        ----------
        db: Connection
            DB connection.
        record: cls | {str: object}
            Model object or dictionary of columns and values.
        qualifier: {str: str -> str}
            Functions converting place holders.

        Returns
        -------
        Model
            Model object.
        """
        record = record if isinstance(record, cls) else cls(**record)
        value_dict = model_values(cls, record)
        check_columns(cls, value_dict)
        cols, vals = list(value_dict.keys()), list(value_dict.values())
        qualifier = key_to_index(qualifier, cols)

        db.stmt().execute(f"INSERT INTO {cls.name} ({', '.join(cols)}) VALUES {values(len(cols), 1, qualifier)}", *vals)

        for c, v in cls.last_sequences(db, 1):
            setattr(record, c.name, v)

        return record

    @classmethod
    def update(cls, db, pks, values, qualifier={}):
        """
        Update a record by primary key(s).

        This method only updates columns which are found in `values` and not primary key.

        Parameters
        ----------
        db: Connection
            DB connection.
        pks: object | {str: object}
            Representation of primary key(s).
        values: cls | {str: object}
            Model object or dictionary of columns and values.
        qualifier: {str: str -> str}
            Functions converting place holders.

        Returns
        -------
        bool
            Boolean which shows whether a row is updated or not. When there exists no row having given primary key(s), `False` is returned.
        """
        cols, vals = parse_pks(cls, pks)
        return cls.update_where(db, values, Conditional.all([Q.eq(**{c: v}) for c, v in zip(cols, vals)]), qualifier) == 1

    @classmethod
    def update_where(cls, db, values, condition, qualifier={}, allow_all=True):
        """
        Update records which satisfy a condition.

        Parameters
        ----------
        db: Connection
            DB connection.
        values: cls | {str: object}
            Model object or dictionary of columns and values.
        condition: Conditional
            Condition.
        qualifier: {str: str -> str}
            Functions converting place holders.
        allow_all: bool
            Empty condition raises an exception if this is `False`.

        Returns
        -------
        int
            The number of affected rows.
        """
        value_dict = model_values(cls, values, excludes_pk=True)
        check_columns(cls, value_dict)
        cols, vals = list(value_dict.keys()), list(value_dict.values())
        qualifier = key_to_index(qualifier, cols)

        def set_col(acc, icv):
            i, (c, v) = icv
            if isinstance(v, Expression):
                clause = f"{c} = {qualifier.get(i, lambda x:x)(v.expression)}"
                params = v.params
            else:
                clause = f"{c} = {qualifier.get(i, lambda x:x)('$_')}"
                params = [v]
            acc[0].append(clause)
            acc[1].extend(params)
            return acc

        setters, params = reduce(set_col, enumerate(zip(cols, vals)), ([], []))

        wc, wp = where(condition)
        if wc == "" and not allow_all:
            raise ValueError("Update query to update all records is not allowed.")

        c = db.stmt().execute(f"UPDATE {cls.name} SET {', '.join(setters)}{_spacer(wc)}", *(params + wp))

        return getattr(c, "rowcount", None)

    @classmethod
    def delete(cls, db, pks):
        """
        Delete a record by primary key(s).

        Parameters
        ----------
        db: Connection
            DB connection.
        pks: object | {str: object}
            Representation of primary key(s).

        Returns
        -------
        bool
            Boolean which shows whether a row is deleted or not. When there exists no row having given primary key(s), `False` is returned.
        """
        cols, vals = parse_pks(cls, pks)
        return cls.delete_where(db, Conditional.all([Q.eq(**{c: v}) for c, v in zip(cols, vals)])) == 1

    @classmethod
    def delete_where(cls, db, condition, allow_all=True):
        """
        Delete records which fulfill a condition.

        Parameters
        ----------
        db: Connection
            DB connection.
        condition: Conditional
            Condition.
        allow_all: bool
            Empty condition raises an exception if this is `False`.

        Returns
        -------
        int
            The number of affected rows.
        """
        wc, wp = where(condition)
        if wc == "" and not allow_all:
            raise ValueError("Delete query to delete all records is not allowed.")

        c = db.stmt().execute(f"DELETE FROM {cls.name}{_spacer(wc)}", *wp)

        return getattr(c, "rowcount", None)

    @classmethod
    def last_sequences(cls, db, num):
        """
        Returns latest auto generated numbers in this table.

        Parameters
        ----------
        db: Connection
            DB connection.
        num: int
            The number of records inserted by the latest insert query.

        Returns
        -------
        [(Column, int)]
            A list of pairs of column and the generated number.
        """
        return []


def _spacer(s):
    return (" " + str(s)) if s else ""


import sys
import types
from pyracmon.config import pyracmon, default_config
from pyracmon.connection import connect, Connection
from pyracmon.context import ConnectionContext
from pyracmon.mixin import CRUDMixin
from pyracmon.select import read_row
from pyracmon.model import define_model, Table, Column
from pyracmon.model_graph import GraphEntityMixin
from pyracmon.query import Q, Conditional, where
from pyracmon.graph import new_graph, GraphSpec, S
from pyracmon.graph.schema import TypedDict, document_type, Typeable, walk_schema, GraphSchema


__all__ = [
    "pyracmon",
    "connect",
    "Connection",
    "ConnectionContext",
    "CRUDMixin",
    "read_row",
    "define_model",
    "Table",
    "Column",
    "Q",
    "Conditional",
    "where",
    "new_graph",
    "S",
    "document_type",
    "Typeable",
    "walk_schema",
    "GraphSchema",
    "declare_models",
    "graph_template",
    "graph_dict",
    "graph_schema",
]

if sys.version_info[0:2] < (3, 8):
    __all__ += ["TypedDict"]


def declare_models(dialect, db, module=__name__, mixins=[], excludes=None, includes=None):
    """
    Declare model types read from database in the specified module.

    Parameters
    ----------
    dialect: module
        A module exporting `read_schema` function and `mixins` classes.
    db: pyracmon.connection.Connection
        Wrapper of DB-API 2.0 Connection.
    module: str | module
        A module name where the declarations are located.
    mixins: [type]
        Additional mixin classes for declaring model types.
    excludes: [str]
        Excluding table names.
    includes: [str]
        Including table names. All tables excluding `excludes` are declared as models if this argument is omitted.
    """
    tables = dialect.read_schema(db, excludes, includes)
    for t in tables:
        if isinstance(module, types.ModuleType):
            module.__dict__[t.name] = define_model(t, mixins + dialect.mixins + [CRUDMixin, GraphEntityMixin])
        else:
            sys.modules[module].__dict__[t.name] = define_model(t, mixins + dialect.mixins + [CRUDMixin, GraphEntityMixin])


def graph_template(*bases, **definitions):
    """
    Create a graph template on the default specification predefined to handle model object in appropriate ways.

    Parameters
    ----------
    bases: [GraphTemplate]
        Base templates whose properties and relations are merged into new template.
    template: {str: (type | Tuple[type, T -> ID, T -> bool])}
        Definitions of template properties.

    Returns
    -------
    GraphTemplate
        Created graph template.
    """
    return default_config().graph_spec.new_template(*bases, **definitions)


def graph_dict(graph, **settings):
    """
    Generates a dictionary representing structured values of a graph on the default specification.

    Parameters
    ----------
    graph: GraphView
        A view of the graph.
    settings: {str: NodeSerializer}
        Mapping from property name to `NodeSerializer` s.

    Returns
    -------
    {str: object}
        A dictionary representing the graph.
    """
    return default_config().graph_spec.to_dict(graph, **settings)


def graph_schema(template, **settings):
    """
    Creates `GraphSchema` under the default specifications.

    Parameters
    ----------
    template: GraphTemplate
        A template of serializing graph.
    settings: {str: NodeSerializer}
        Mapping from property name to `NodeSerializer` s.

    Returns
    -------
    GraphSchema
        An object having schema information of serialized dictionary.
    """
    return default_config().graph_spec.to_schema(template, **settings)


import string
import threading
from datetime import datetime
from .query import QueryHelper
from .sql import Sql
from .marker import Marker
from .context import ConnectionContext


def connect(api, *args, **kwargs):
    """
    Connect to DB by passing arguments to DB-API 2.0 module.

    Every optional argument is passed to `api.connect()` and returns the `Connection` object which wraps obtained DB connection.

    Parameters
    ----------
    api: module
        DB-API 2.0 module which should export `connect()` function.

    Returns
    -------
    Connection
        Wrapper of DB-API 2.0 connection.
    """
    return Connection(api, api.connect(*args, **kwargs), None)


class Connection:
    """
    Wrapper class of DB-API 2.0 Connection.

    Every instance works as the proxy object to original connection, therefore any attribute in it is still available.
    """
    _characters = string.ascii_letters + string.digits + ".="

    def __init__(self, api, conn, context_factory=None):
        self.identifier = self._gen_identifier()
        self.api = api
        self.conn = conn
        self.context_factory = context_factory
        self._context = None

    def __getattr__(self, name):
        return getattr(self.conn, name)

    def __enter__(self):
        if hasattr(self.conn, "__enter__"):
            self.conn.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if hasattr(self.conn, "__exit__"):
            self.conn.__exit__(exc_type, exc_value, traceback)
        else:
            if exc_value is None:
                self.conn.rollback()
            else:
                self.conn.commit()
            self.conn.close()

    def _gen_identifier(self):
        t = int(datetime.now().timestamp() * 1000)

        def gen(x):
            base = len(Connection._characters)
            while x >= base:
                x, r = divmod(x, base)
                yield Connection._characters[r]
            yield Connection._characters[x]

        return threading.current_thread().name + "-" + ''.join(gen(t))[::-1]

    @property
    def helper(self):
        """
        Deprecated. Returns `QueryHelper` object.
        """
        return QueryHelper(self.api, None)

    @property
    def context(self):
        """
        `ConnectionContext` used for this connection.

        Returns
        -------
        ConnectionContext
            `ConnectionContext` used for this connection.
        """
        if not self._context:
            self._context = (self.context_factory or ConnectionContext)()
            self._context.identifier = self.identifier
        return self._context

    def use(self, factory):
        """
        Set factory function of `ConnectionContext`.

        Parameters
        ----------
        factory: () -> ConnectionContext
            A factory function.

        Returns
        -------
        Connection
            This instance.
        """
        self.context_factory = factory
        return self

    def stmt(self, context=None):
        """
        Creates new statement which provides methods to execute query.

        Parameters
        ----------
        context: ConnectionContext
            `ConnectionContext` used in the statement. If `None`, the context of this connection is used.

        Returns
        -------
        Statement
            Created statement.
        """
        return Statement(self, context or self.context)


class Statement:
    """
    This class has the functionality to execute query on containing connection and context.
    """
    def __init__(self, conn, context):
        self.conn = conn
        self.context = context

    def prepare(self, sql, *args, **kwargs):
        """
        Generates a prepared SQL statement and its parameters.

        Parameters
        ----------
        sql: str
            SQL template.
        args: [object]
            Indexed parameters in the SQL.
        kwargs: {str: object}
            Keyword parameters in the SQL.

        Returns
        -------
        str
            Prepared SQL statement.
        [object] | {str:object}
            Paremeters for created statement.
        """
        paramstyle = self.context.config.paramstyle or self.conn.api.paramstyle

        sql = Sql(Marker.of(paramstyle), sql)

        return sql.render(*args, **kwargs)

    def execute(self, sql, *args, **kwargs):
        """
        Executes a query and returns a cursor object used for the execution.

        Parameters
        ----------
        sql: str
            SQL template.
        args: [object]
            Indexed parameters in the SQL.
        kwargs: {str: object}
            Keyword parameters in the SQL.

        Returns
        -------
        Cursor
            Cursor object which has been used for the query execution.
        """
        sql, params = self.prepare(sql, *args, **kwargs)

        c = self.conn.cursor()

        return self.context.execute(c, sql, params)


from functools import partial, reduce
from inspect import signature, Signature
from typing import TypeVar


T = TypeVar('T')


def as_is(x):
    return x


def wrap_serializer(f):
    try:
        sig = signature(f)
        def g(cxt, node, base, value) -> sig.return_annotation:
            ba = sig.bind(*(value, base, node, cxt)[len(sig.parameters)-1::-1])
            return f(*ba.args)
        return g
    except:
        def g(cxt, node, base, value):
            return f(value)
        return g


def chain_serializers(serializers):
    def merge(fs):
        rt = Signature.empty
        for f in fs[::-1]:
            t = signature(f).return_annotation
            if t != Signature.empty:
                try:
                    t[T]
                    rt = t if rt == Signature.empty else rt[t]
                except TypeError:
                    try:
                        return rt[t]
                    except TypeError:
                        return t
        return rt

    funcs = [wrap_serializer(f) for f in serializers]
    rt = merge(funcs)

    def composed(cxt, node, base, value) -> rt:
        base = [as_is, wrap_serializer(base)] if base else [as_is]
        return reduce(lambda acc,f: partial(f, cxt, node, acc), base + funcs)(value)

    return composed


from itertools import zip_longest
from functools import reduce, partial
from pyracmon.graph.identify import IdentifyPolicy, HierarchicalPolicy, neverPolicy
from pyracmon.graph.template import GraphTemplate
from pyracmon.graph.serialize import S, SerializationContext, NodeSerializer
from pyracmon.graph.schema import GraphSchema
from pyracmon.util import Configurable


class GraphSpec:
    """
    This class contains the specifications of graph which control various behaviors in the lifecycles of graphs.

    Each instance contains 3 kind of functions; *identifier*, *entity filter* and *serializer*.

    *Identifier* is a function to get a value used for the identification of graph entity. See `Graph` to know how this works.

    *Entity fliter* is a function returning `bool` to determine whether the entity should be appended to a graph or not.
    If `False` is returned for an entity, it is just ignored.

    *Serializer* is a function which converts an entity value into a serializable object.
    This is used in `to_dict()` to put entities into returning `dict`.
    Registered *serializer* is first applied to an entity value and other serialization functions in `NodeSerializer` are applied in order.

    They are registered in the specification via `add_xxx()` methods of this class with its bound `type`.
    The first function whose bound `type` is the subclass of the `kind` of the template property is choosed and applied.

    Attributes
    ----------
    identifiers: [(type, T -> ID)]
        A list of pairs of type and *identifier*.
    entity_filters: [(type, T -> bool)]
        A list of pairs of type and *entity filter*.
    serializers: [(type, T -> U)]
        A list of pairs of type and *serializer*.
    """
    def __init__(self, identifiers=None, entity_filters=None, serializers=None):
        self.identifiers = identifiers or []
        self.entity_filters = entity_filters or []
        self.serializers = serializers or []

    def _get_inherited(self, holder, t):
        if not isinstance(t, type):
            return None
        return next(map(lambda x:x[1], filter(lambda x:issubclass(t, x[0]), holder)), None)

    def get_identifier(self, t):
        return self._get_inherited(self.identifiers, t)

    def get_entity_filter(self, t):
        return self._get_inherited(self.entity_filters, t)

    def find_serializers(self, t):
        if not isinstance(t, type):
            return []
        return list(map(lambda x:x[1], filter(lambda x:issubclass(t, x[0]), self.serializers[::-1])))

    def add_identifier(self, c, f):
        """
        Register an *identifier* with the bound type.

        Parameters
        ----------
        c: type
            A type bound for the *identifier*.
        f: T -> ID
            An *identifier* function.

        Returns
        -------
        GraphSpec
            This instance.
        """
        self.identifiers[0:0] = [(c, f)]
        return self

    def add_entity_filter(self, c, f):
        """
        Register an *entity filter* with the bound type.

        Parameters
        ----------
        c: type
            A type bound for the *entity filter*.
        f: T -> bool
            An *entity filter* function.

        Returns
        -------
        GraphSpec
            This instance.
        """
        self.entity_filters[0:0] = [(c, f)]
        return self

    def add_serializer(self, c, f):
        """
        Register a *serializer* with the bound type.

        Parameters
        ----------
        c: type
            A type bound for the *serializer*.
        f: T -> U
            An *serializer* function.

        Returns
        -------
        GraphSpec
            This instance.
        """
        if isinstance(f, NodeSerializer):
            f = f.serializer
        self.serializers[0:0] = [(c, f)]
        return self

    def _make_policy(self, t, f):
        f = f or self.get_identifier(t)

        if isinstance(f, IdentifyPolicy):
            return f
        elif callable(f):
            return HierarchicalPolicy(f)
        else:
            return neverPolicy()

    def _get_property_definition(self, d):
        if d is None or isinstance(d, tuple):
            d = iter(d or ())
            kind = next(d, None)
            ident = self._make_policy(kind, next(d, None))
            ef = next(d, self.get_entity_filter(kind))
            return kind, ident, ef
        elif isinstance(d, type):
            return d, self._make_policy(d, None), self.get_entity_filter(d)
        elif isinstance(d, GraphTemplate):
            return d, neverPolicy(), None
        else:
            raise ValueError(f"Invalid value was found in keyword arguments of new_template().")

    def new_template(self, *bases, **template):
        """
        Creates a graph template with given definitions for template properties.

        Each keyword argument corresponds to a template property where the key is proprety name and value is property definition.

        Property definition can be given as a single value or a tuple of at most 3 values.
        The values are interpreted to property attributes in the following order.

        - The kind of entity value which must be a `type` object.
        - *Identifier* of the property.
        - *Entity filter* of the property.

        Parameters
        ----------
        bases: [GraphTemplate]
            Base templates whose properties and relations are merged into new template.
        template: {str: (type | Tuple[type, T -> ID, T -> bool])}
            Definitions of template properties.

        Returns
        -------
        GraphTemplate
            Created graph template.
        """
        base = sum(bases, GraphTemplate([]))

        return base + GraphTemplate([(n, *self._get_property_definition(d)) for n, d in template.items()])

    def to_dict(self, graph, __params={}, **settings):
        """
        Generates a dictionary representing structured entity values of a graph.

        Only nodes whose names appear in keys of `settings` are contained int the result.

        `NodeSerializer` object of each value in `settings` keyword arguments can be built by builder methods of `S`.

        Parameters
        ----------
        graph: GraphView
            A view of the graph.
        __params: {str: {str: object}}
            Parameters keyed by node names.
        settings: {str: NodeSerializer}
            Mapping from property name to `NodeSerializer` s.

        Returns
        -------
        Dict[str, object]
            A dictionary representing the graph.
        """
        return SerializationContext(settings, self.find_serializers, __params).execute(graph)

    def to_schema(self, template, **settings):
        """
        Creates `GraphSchema` under this specifications.

        Parameters
        ----------
        template: GraphTemplate
            A template of serializing graph.
        settings: {str: NodeSerializer}
            Mapping from property name to `NodeSerializer` s.

        Returns
        -------
        GraphSchema
            An object having schema information of serialized dictionary.
        """
        return GraphSchema(self, template, **settings)


class IdentifyPolicy:
    """
    Provides functionalities to handle identical entities in appending sessions of a graph.
    """
    def __init__(self, identifier):
        self.identifier = identifier

    def get_identifier(self, value):
        """
        Returns identification key from an entity value.

        Parameters
        ----------
        value: object
            An entity value.

        Returns
        -------
        object
            Identification key of the entity value.
        """
        return self.identifier(value) if self.identifier else None

    def identify(self, prop, candidates, ancestors):
        """
        Select parent nodes and identical nodes of a new entity.

        Parameters
        ----------
        prop: GraphTemplate.Property
            Template property for new entity.
        candidates: [Node]
            Nodes having the same identification key.
        ancestors: {str: [Node]}
            Mappings from property name to identical node set which appeared in current appending session.

        Returns
        -------
        [Node | None]
            Parent nodes which the node of new entity should be appended newly.
            `None` means to append a new node without parent.
        [Node]
            Identical existing nodes propagated to identifications of children as ancestors.
        """
        raise NotImplementedError()


class HierarchicalPolicy(IdentifyPolicy):
    """
    Default policy for template properties having identifier.

    This policy identifies nodes whose entity has the same identification key as one of appending entity
    and whose parent is also identical to the parent of the entity.
    """
    def identify(self, prop, candidates, ancestors):
        if prop.parent and prop.parent.name in ancestors:
            # This entity has parent in this session.
            parents = ancestors[prop.parent.name]

            parent_nodes = []

            for p in parents:
                # Find parent nodes which don't have child of the same identifier.
                if all([not p.children[prop.name].has(n) for n in candidates]):
                    parent_nodes.append(p)

            # Find identical nodes from candidates by checking whether the node belogs to a parent contained in ancestors.
            identical_nodes = [n for n in candidates if any([p.children[prop.name].has(n) for p in parents])]

            return parent_nodes, identical_nodes
        else:
            # This entity is a root entity. Key equality is only a criteria of identification.
            return ([], candidates) if candidates else ([None], [])


class NeverPolicy(IdentifyPolicy):
    def identify(self, prop, candidates, ancestors):
        if prop.parent and prop.parent.name in ancestors:
            return ancestors[prop.parent.name], []
        else:
            return [None], []


def neverPolicy(instance=NeverPolicy(None)):
    return instance


class AlwaysPolicy(IdentifyPolicy):
    def identify(self, prop, candidates, ancestors):
        if prop.parent and prop.parent.name in ancestors:
            parents = ancestors[prop.parent.name]

            parent_nodes = []

            for p in parents:
                # Find parent nodes which don't have child of the same identifier.
                if all([not p.children[prop.name].has(n) for n in candidates]):
                    parent_nodes.append(p)

            return parent_nodes, candidates
        else:
            return ([], candidates) if candidates else ([None], [])


import re
from decimal import Decimal
from datetime import date, datetime, time, timedelta
from uuid import UUID
from itertools import groupby
from pyracmon.model import Table, Column
from pyracmon.dialect.shared import MultiInsertMixin
from pyracmon.query import Q, where, holders


SequencePattern = re.compile(r"nextval\(\'([a-zA-Z0-9_]+)\'(\:\:regclass)?\)")


def read_schema(db, excludes=None, includes=None):
    """
    Collect tables in current database.

    Parameters
    ----------
    excludes: [str]
        Excluding table names.
    includes: [str]
        Including table names. If not specified, all tables are collected.

    Returns
    -------
    [Table]
        Tables.
    """
    q = Q(excludes = excludes, includes = includes)

    cond = Q.of("c.table_catalog = current_catalog") & Q.eq("c", table_schema="public") & Q.in_("t", table_type=["BASE TABLE", "VIEW"]) \
        & q.excludes.not_in("c.table_name") & q.includes.in_("c.table_name")

    w, params = where(cond)

    cursor = db.stmt().execute(f"""\
        SELECT
            c.table_name, c.column_name, c.data_type, c.udt_name,
            e.data_type, e.udt_name, k.constraint_type, c.column_default, c.ordinal_position
        FROM
            information_schema.columns AS c
            INNER JOIN information_schema.tables AS t ON c.table_name = t.table_name
            LEFT JOIN (
                SELECT
                    tc.table_name, k.column_name, string_agg(tc.constraint_type, ',') AS constraint_type
                FROM
                    information_schema.key_column_usage AS k
                    INNER JOIN information_schema.table_constraints AS tc ON k.constraint_name = tc.constraint_name
                WHERE
                    tc.constraint_type = 'PRIMARY KEY' OR tc.constraint_type = 'FOREIGN KEY'
                GROUP BY
                    tc.table_name, k.column_name
            ) AS k ON t.table_name = k.table_name AND c.column_name = k.column_name
            LEFT JOIN information_schema.element_types AS e
                ON ((c.table_catalog, c.table_schema, c.table_name, 'TABLE', c.dtd_identifier) =
                    (e.object_catalog, e.object_schema, e.object_name, e.object_type, e.collection_type_identifier))
        {w}
        ORDER BY c.table_name ASC, c.ordinal_position ASC
        """, *params)

    def map_types(t, udt):
        base = db.context.config.type_mapping
        ptype = base and base(t, udt_name=udt)
        return ptype or _map_types(t)

    def column_of(n, t, udt, et, eudt, constraint, default, pos):
        m = SequencePattern.match(default or "")
        cs = (constraint or "").split(',')
        seq = m.group(1) if m else None
        ptype = map_types(t, udt) if t != 'ARRAY' else [map_types(et, eudt)]
        info = (t, udt) if t != 'ARRAY' else (et, eudt)
        return Column(n, ptype, info, 'PRIMARY KEY' in cs, 'FOREIGN KEY' in cs, seq)

    tables = []
    column_positions = {}

    for t, cols in groupby(cursor.fetchall(), lambda row: row[0]):
        cols = list(cols)
        columns = [column_of(*c[1:]) for c in cols]
        tables.append(Table(t, columns))
        column_positions[t] = {c[1]:c[-1] for c in cols}

    cursor.close()

    # Materialized views
    cond = Q.eq("c", relkind = "m") & Q.ge("a", attnum = 1) \
        & q.excludes.not_in("c.relname") & q.includes.in_("c.relname")

    w, params = where(cond)

    cursor = db.stmt().execute(f"""\
        SELECT
            c.relname, a.attname, t.typname, et.typname, a.attnum
        FROM
            pg_class AS c
            INNER JOIN pg_attribute AS a ON c.oid = a.attrelid
            INNER JOIN pg_type AS t ON a.atttypid = t.oid
            LEFT JOIN pg_type AS et ON t.typelem = et.oid
        {w}
        ORDER BY
            c.oid ASC, a.attnum ASC
        """, *params)

    def mv_column_of(n, udt, eudt, pos):
        ptype = map_types(_map_alternates(udt), udt) if eudt is None else [map_types(_map_alternates(eudt), eudt)]
        info = (_map_alternates(udt), udt) if eudt is None else (_map_alternates(eudt), eudt)
        return Column(n, ptype, info, False, False, None)

    for t, cols in groupby(cursor.fetchall(), lambda row: row[0]):
        cols = list(cols)
        columns = [mv_column_of(*c[1:]) for c in cols]
        tables.append(Table(t, columns))
        column_positions[t] = {c[1]:c[-1] for c in cols}

    cursor.close()

    if len(tables) == 0:
        return tables

    cursor = db.stmt().execute(f"""\
        SELECT
            relname, oid
        FROM
            pg_class
        WHERE
            relname IN ({holders(len(tables))})
        """, *[t.name for t in tables])

    table_oids = {}
    for n, oid in cursor.fetchall():
        table_oids[n] = oid

    for t in tables:
        cc = db.stmt().execute(f"SELECT col_description($_, 0)", *[table_oids[t.name]])
        t.comment = cc.fetchone()[0] or ""

        for i, col in enumerate(t.columns):
            m = db.helper.marker()
            cc = db.stmt().execute(f"SELECT col_description($_, $_)", *[table_oids[t.name], column_positions[t.name][col.name]])
            col.comment = cc.fetchone()[0] or ""

        cc.close()

    cursor.close()

    return tables


def _map_types(t):
    if t == "boolean":
        return bool
    elif t == "real" or t == "double precision":
        return float
    elif t == "smallint" or t == "integer" or t == "bigint":
        return int
    elif t == "numeric" or t == "decimal":
        return Decimal
    elif t == "character varying" or t == "text" or t == "character":
        return str
    elif t == "bytea":
        return bytes
    elif t == "date":
        return date
    elif t.startswith("timestamp "):
        return datetime
    elif t == "time" or t.startswith("time "):
        return time
    elif t == "interval":
        return timedelta
    elif t == "uuid":
        return UUID
    else:
        return object


def _map_alternates(n):
    if n == "int2":
        return "smallint"
    elif n == "int" or n == "int4":
        return "integer"
    elif n == "int8":
        return "bigint"
    elif n == "float4":
        return "real"
    elif n == "float8":
        return "double precision"
    elif n == "decimal":
        return "numeric"
    elif n == "bool":
        return "boolean"
    elif n == "char":
        return "character"
    elif n == "varchar":
        return "character varying"
    elif n == "timetz":
        return "time with time zone"
    elif n == "timestamptz":
        return "timestamp with time zone"
    else:
        return n


class PostgreSQLMixin(MultiInsertMixin):
    @classmethod
    def last_sequences(cls, db, num):
        cols = [c for c in cls.columns if c.incremental]

        if len(cols) > 0:
            sequences = []
            d = db.cursor()
            for c in cols:
                d.execute(f"SELECT currval('{c.incremental}')")
                sequences.append((c, d.fetchone()[0]))
            d.close()
            return sequences
        else:
            return []

    @classmethod
    def truncate(cls, db):
        db.cursor().execute(f"TRUNCATE {cls.name} RESTART IDENTITY CASCADE")


mixins = [PostgreSQLMixin]


class Selection:
    """
    The representation of table and its columns used in SQL.

    This class is designed to be a bridge from query generation to reading results.
    String expression of the instance is comma-separated column names with alias, which can be embedded in the select query.

    >>> s1 = table1.select("t1", includes = ["col11", "col12"])
    >>> s2 = table2.select("t2")
    >>> str(s1)
    't1.col11, t1.col12'
    >>> str(s2)
    't2.col21, t2.col22, t2.col23'

    The instances of this class are used in `read_row()` to read model object from obtained row.

    >>> c.execute(f"SELECT {s1}, {s2} FROM table1 AS t1 INNER JOIN table2 AS t2 ON ...")
    >>> for row in c.fetchall():
    >>>     r = read_row(row, s1, s2)
    >>>     assert isinstance(r.t1, table1)
    >>>     assert isinstance(r.t2, table2)
    """
    def __init__(self, table, alias, columns):
        self.table = table
        self.alias = alias
        self.columns = columns

    @property
    def name(self):
        return self.alias if self.alias else self.table.name

    def __len__(self):
        return len(self.columns)

    def __repr__(self):
        a = f"{self.alias}." if self.alias else ""
        return ', '.join([f"{a}{c.name}" for c in self.columns])

    def __add__(self, other):
        return FieldExpressions() + self + other

    def __iter__(self):
        return iter([self])

    def consume(self, values):
        return self.table(**dict([(c.name, v) for c, v in zip(self.columns, values)]))


class FieldExpressions:
    """
    The instance of this class works as the composition of `Selection` s which provides attributes to access each `Selection`.

    `+` operation on `Selection` s creates an instance of `FieldExpressions`. Each `Selection` is available via attributes of its name.

    >>> exp = table1.select("t1", includes=["col11", "col12"]) + table2.select("t2")
    >>> c.execute(f"SELECT {exp} FROM table1 AS t1 INNER JOIN table2 AS t2 ON ...")
    >>> for row in c.fetchall():
    >>>     r = read_row(row, *exp)
    >>>     assert isinstance(r.t1, table1)
    >>>     assert isinstance(r.t2, table2)

    Empty tuple and string are available in addition to `Selection` object.
    They are replaced with index arguments and keywords arguments each other on the direct invocation of the instance.

    >>> exp = table1.select("t1", includes=["col11", "col12"]) + () + "a" + () + "b"
    >>> f"{exp("t2.col21", "t2.col23", a="t2.col22", b="t2.col24")}"
    t1.col11, t1.col12, t2.col21, t2.col22, t2.col23, t2.col24
    """
    def __init__(self):
        self.__selections = []
        self.__keys = {}

    def __add__(self, other):
        exp = FieldExpressions()
        exp += self
        exp += other
        return exp

    def __iadd__(self, other):
        if isinstance(other, Selection):
            self.__selections.append(other)
            self.__keys[other.name] = other
        elif isinstance(other, FieldExpressions):
            self.__selections += other.__selections
            self.__keys.update(other.__keys)
        elif isinstance(other, str):
            self.__selections.append(other)
            self.__keys[other] = other
        elif other is ():
            self.__selections.append(other)
        else:
            raise ValueError(f"Operand of + for FieldExpressions must be a Selection or FieldExpressions but {type(other)} is given.")
        return self

    def __getattr__(self, key):
        return self.__keys[key]

    def __iter__(self):
        return iter(self.__selections)

    class Instance:
        def __init__(self, exp, *args, **kwargs):
            self.exp = exp
            self.args = args
            self.kwargs = kwargs

        def __repr__(self):
            args = list(self.args)
            def _repr(s):
                if isinstance(s, Selection):
                    return s.__repr__()
                elif isinstance(s, str):
                    return self.kwargs.get(s, s)
                elif s == ():
                    return args.pop(0)
            return ', '.join(map(_repr, self.exp))

    def __call__(self, *args, **kwargs):
        return FieldExpressions.Instance(self, *args, **kwargs)

    def __repr__(self):
        return self().__repr__()


class RowValues:
    """
    This class provides attribute access for each row in query result.

    Each instance behaves as if it is a list of values created by `Selection` s.
    Index access returns the value at the index and iteration yields values in order.

    >>> exp = table1.select("t1"), table2.select()
    >>> r = read_row(row, *exp)
    >>> r[0]
    ...
    >>> [v for v in r]
    ...

    It also exposes attributes whose name is the alias (if exists) or the table name.

    >>> r.t1
    ...
    >>> r.table2
    ...
    """
    def __init__(self, selections):
        self.key_map = dict([(s, i) for i, s in enumerate(map(self._key_of, selections)) if s is not None])
        self.values = []

    def _key_of(self, s):
        if isinstance(s, Selection):
            return s.name
        elif isinstance(s, str):
            return s
        else:
            return None

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def __getitem__(self, index):
        return self.values[index]

    def __getattr__(self, key):
        index = self.key_map.get(key, None)
        if index is None:
            raise AttributeError(f"No selection is found whose table name or alias is '{key}'")
        return self.values[index]

    def append(self, value):
        self.values.append(value)


def read_row(row, *selections, allow_redundancy=False):
    """
    Read values in a row according to `selections`.

    This function returns `RowValues` where each value is created by the item of `selections` respectively.
    The type of the item determines how values in the row is handled:

    - Selection consumes as many values as the number of columns in it and creates a model instance.
    - Callable consumes a value and returns another value.
    - Empty tuple or a string consumes a value, which is stored in `RowValues` as it is.

    Parameters
    ----------
    row: object
        An object representing a row returned by fetchone() / fetchall().
    selections: [Selection | S -> T | () | str]
        various type of objects determining the way to handle values in the row.
    allow_redundancy: bool
        If `False`, an exception is thrown when not all values in a row are consumed.

    Returns
    -------
    RowValues
        Values read from the row.
    """
    result = RowValues(selections)

    for s in selections:
        if isinstance(s, Selection):
            result.append(s.consume(row))
            row = row[len(s):]
        elif callable(s):
            result.append(s(row[0]))
            row = row[1:]
        elif s == () or isinstance(s, str):
            result.append(row[0])
            row = row[1:]
        else:
            raise ValueError("Unavailable value is given to read_row().")

    if not allow_redundancy and len(row) > 0:
        raise ValueError("Not all elements in row is consumed.")

    return result


class SelectMixin:
    @classmethod
    def select(cls, alias="", includes=[], excludes=[]):
        """
        Select columns to use in a query with an alias of this table.

        Parameters
        ----------
        alias: str
            An alias string of this table.
        includes: [str]
            Column names to use. All columns are selected if empty.
        excludes: [str]
            Column names not to use.

        Returns
        -------
        Selection
            An object which has selected columns.
        """
        columns = [c for c in cls.columns if c.name not in excludes] \
            if not bool(includes) else \
                [c for c in cls.columns if c.name not in excludes and c.name in includes]
        return Selection(cls, alias, columns)



from typing import TypeVar
from itertools import takewhile
import inspect
from .util import Configurable
from .graph.spec import GraphSpec
from .graph.schema import TypedDict, DynamicType, Shrink, issubgeneric, document_type
from .graph.serialize import T
from .graph.util import chain_serializers


class GraphEntityMixin:
    """
    Mixin class for model types which enables identity calculation and nullity check.
    """
    @classmethod
    def identity(cls, model):
        """
        Returns primary key values as the identity of a model.

        Parameters
        ----------
        model: Model
            A model object.

        Returns
        -------
        tuple
            Primary keys as the identity of the model. If the model type has no primary key, returns `None`.
        """
        pks = [c.name for c in cls.columns if c.pk]
        if len(pks) > 0 and all([hasattr(model, n) and getattr(model, n) is not None for n in pks]):
            return tuple(map(lambda n: getattr(model, n), pks))
        else:
            return None

    @classmethod
    def is_null(cls, model):
        """
        Checks the model is considered to be null.

        Parameters
        ----------
        model: Model
            A model object.

        Returns
        -------
        bool
            `True` if all column values set to the model are `None`, otherwise `False`.
        """
        return all([getattr(model, c.name, None) is None for c in cls.columns])


class ModelSchema(DynamicType[T]):
    @classmethod
    def fix(cls, bound, arg):
        class Schema(TypedDict):
            pass
        setattr(Schema, '__annotations__', {c.name:document_type(c.ptype, c.comment) for c in bound.columns})
        return Schema


class ExcludeFK(Shrink[T]):
    @classmethod
    def select(cls, bound, arg):
        return {c.name for c in arg.columns if c.fk}, None


class ConfigurableSpec(GraphSpec, Configurable):
    """
    Extension of `GraphSpec` prepared to integrate model types into graph specification.

    This class exposes additional configurable attributes which controls the graph operation for model types.
    """
    @classmethod
    def create(cls):
        spec = cls()

        spec.add_identifier(GraphEntityMixin, lambda m: type(m).identity(m))
        spec.add_entity_filter(GraphEntityMixin, lambda m: m and not type(m).is_null(m))

        def serialize(model:T) -> ModelSchema[T]:
            return {c.name:v for c, v in model}
        spec.add_serializer(GraphEntityMixin, serialize)

        return spec

    def __init__(self, *args):
        super(ConfigurableSpec, self).__init__(*args)

        self.include_fk = False

    def clone(self):
        spec = ConfigurableSpec(
            self.identifiers.copy(),
            self.entity_filters.copy(),
            self.serializers.copy(),
        )
        spec.include_fk = self.include_fk
        return spec

    def replace(self, another):
        self.identifiers[:] = another.identifiers
        self.entity_filters[:] = another.entity_filters
        self.serializers[:] = another.serializers
        self.include_fk = another.include_fk

    def _model_serializer(self, bases):
        """
        Generate configured serializer for model type.

        Parameters
        ----------
        bases: [Model -> object]
            Serialization functions.
        """
        if not self.include_fk:
            def serialize(c, n, b, model:T) -> ExcludeFK[T]:
                d = {c.name:v for c, v in model if not c.fk}
                return b(type(model)(**d))

            pos = next(filter(lambda ib: issubgeneric(inspect.signature(ib[1]).return_annotation, ModelSchema), enumerate(bases)), None)

            return bases[0:pos[0]+1] + [serialize] + bases[pos[0]+1:] if pos else bases
        else:
            return bases

    def find_serializers(self, t):
        bases = super(ConfigurableSpec, self).find_serializers(t)

        if issubclass(t, GraphEntityMixin):
            return self._model_serializer(bases)
        else:
            return bases




class Marker:
    """
    This class provides the abstration mechanism for marker creation used to embed parameters in a query.

    Each inheriting type corresponds to `paramstyle` defined in DB-API 2.0.

    The instance of `Marker` is callable object designed to return place holder string available in query.
    """
    @classmethod
    def of(cls, paramstyle):
        """
        Select marker type from `paramstyle` and creates its instance.
        """
        if paramstyle == 'qmark':
            return QMarker()
        elif paramstyle == 'numeric':
            return NumericMarker()
        elif paramstyle == 'named':
            return NamedMarker()
        elif paramstyle == 'format':
            return FormatMarker()
        elif paramstyle == 'pyformat':
            return PyformatMarker()
        else:
            raise ValueError(f"Unknown parameter style: {paramstyle}")

    def __call__(self, key=None):
        """
        Renders a place holder string for the key.

        Parameters
        ----------
        key: int | str
            A key specifying the parameter which will be given on SQL execution.

        Returns
        -------
        str
            A place holder string.
        """
        if key is None:
            return self.default()
        elif isinstance(key, int) and key > 0:
            return self.indexed(key)
        elif isinstance(key, str) and key != "":
            return self.keyed(key)
        else:
            raise ValueError("Argument of marker invocation must be a positive int or str.")

    def default(self):
        raise NotImplementedError()

    def indexed(self, index):
        raise NotImplementedError()

    def keyed(self, key):
        raise NotImplementedError()

    def reset(self):
        """
        Resets the internal state.
        """
        pass

    def params(self, *args, **kwargs):
        """
        Creates a holder containing parameters in the form available for the marker type.

        Parameters
        ----------
        args: [object]
            Parameters for indexed place holder.
        kwargs: {str: object}
            Parameters for keyed place holder.

        Returns
        -------
        [object] | {str:object}
            List or dictionary which has parameters at their valid index or key.
        """
        raise NotImplementedError()


class ListMarker(Marker):
    def __init__(self, start=0):
        self.index = start
        self.param_keys = []

    def render(self):
        raise NotImplementedError()

    def default(self):
        self.param_keys.append(self.index)
        self.index += 1
        return self.render()

    def indexed(self, index):
        self.param_keys.append(index-1)
        return self.render()

    def keyed(self, key):
        self.param_keys.append(key)
        return self.render()

    def reset(self):
        self.index = 0
        self.param_keys = []

    def params(self, *args, **kwargs):
        def get(k):
            if isinstance(k, int):
                return args[k]
            else:
                return kwargs[k]

        return [get(k) for k in self.param_keys]


class DictMarker(Marker):
    def __init__(self, start=0):
        self.index = start
        self.param_keys = {}

    def render(self, key):
        raise NotImplementedError()

    def new_key(self, index):
        return f"param{index}"

    def default(self):
        key = self.param_keys.setdefault(self.index, self.new_key(self.index+1))
        self.index += 1
        return self.render(key)

    def indexed(self, index):
        key = self.param_keys.setdefault(index-1, self.new_key(index))
        return self.render(key)

    def keyed(self, key):
        self.param_keys[key] = key
        return self.render(key)

    def reset(self):
        self.index = 0
        self.param_keys = {}

    def params(self, *args, **kwargs):
        def get(k, v):
            if isinstance(k, int):
                return (v, args[k])
            else:
                return (v, kwargs[k])

        return dict([get(k, v) for k, v in self.param_keys.items()])


class QMarker(ListMarker):
    """
    This marker renders `?` for any parameter.
    """
    def render(self):
        return '?'


class NumericMarker(ListMarker):
    """
    This marker renders `:x` for the parameter at index `x` (starting from 1).

    Automatic numbering is used when invoked with no `key`, otherwise specified position is selected.
    The index obtained by next automatic numbering is an incremented value of the last obtained index.
    """
    def render(self):
        return f":{len(self.param_keys)}"


class NamedMarker(DictMarker):
    """
    This marker renders `:key` for the parameter bound to `key`.

    Automatic key selection is used when invoked with no `key`, otherwise specified key is selected.
    Each automatic key is in the form of `keyn` where `n` is specified index or its order in generated keys.
    """
    def render(self, key):
        return f":{key}"


class FormatMarker(ListMarker):
    """
    This marker renders `%s` for any parameter.
    """
    def render(self):
        return '%s'


class PyformatMarker(DictMarker):
    """
    This marker renders `%(key)s` for the parameter bound to `key`.

    This works similarly to `NamedMarker`.
    """
    def render(self, key):
        return f"%({key})s"



class GraphTemplate:
    """
    This class specifies the structure of a graph.

    Each instance is composed of template properties each of which corresponds to a node container of a graph.
    Applying shift operator to properties creates the parent-child relationship between them and it corresponds to an edge of a graph.

    For example, next code creates a template composed of 4 properties in which `d` is a child of `c`, and `b` and `c` are children of `a`.

    >>> template = GraphSpec().new_template(a=int, b=str, c=int, d=float)
    >>> template.a << [template.b, template.c]
    >>> template.c << template.d
    """
    class Property:
        def __init__(self, template, name, kind, policy, entity_filter, origin=None):
            self.template = template
            self.name = name
            self.kind = kind
            self.policy = policy
            self.entity_filter = entity_filter
            self._origin = origin

        def _assert_canbe_parent(self, another):
            if another.parent is not None:
                raise ValueError(f"Graph template property can not have multiple parents.")
            if self.template != another.template:
                raise ValueError(f"Properties can make parent-child relationship only when they are declared in the same template.")
            if self == another:
                raise ValueError(f"Recursive relationship is not allowed.")
            if isinstance(self.kind, GraphTemplate):
                raise ValueError(f"Property for graph template can't have child.")
            p = self
            while p.parent is not None:
                if p.parent == another:
                    raise ValueError(f"Recursive relationship is not allowed.")
                p = p.parent

        @property
        def parent(self):
            """
            Returns parent property.

            Returns
            -------
            GraphTemplate.Property
                Parent property if exists, otherwise `None`.
            """
            return next(filter(lambda r: r[0] == self, self.template._relations), (None, None))[1]

        @property
        def children(self):
            """
            Returns child properties.

            Returns
            -------
            [GraphTemplate.Property]
                Child properties.
            """
            return [r[0] for r in self.template._relations if r[1] == self]

        @property
        def origin(self):
            p = self
            while p._origin:
                p = p._origin
            return p

        def is_compatible(self, other):
            return self.origin is other.origin

        def move_template(self, dest, new_name=None):
            new_name = new_name or self.name
            prop = GraphTemplate.Property(dest, new_name, self.kind, self.policy, self.entity_filter, origin=self)
            _set_template_property(dest, prop)
            for c in self.children:
                cc = c.move_template(dest)
                prop << cc
            return prop

        def __lshift__(self, children):
            """
            Makes this property as a parent of child properties.

            Parameters
            ----------
            children: [GraphTemplate.Property] | GraphTemplate.Property
                Properties to be children of this property.

            Returns
            -------
            [GraphTemplate.Property] | GraphTemplate.Property
                Child properties given in the argument.
            """
            targets = [children] if isinstance(children, GraphTemplate.Property) else children
            for c in targets:
                self._assert_canbe_parent(c)
            self.template._relations += [(c, self) for c in targets]
            return children

        def __rshift__(self, parent):
            """
            Makes this property as a child of another property.

            Parameters
            ----------
            parent: GraphTemplate.Property
                A Property to be a parent of this property.

            Returns
            -------
            GraphTemplate.Property
                A parent property given in `parent` argument.
            """
            parent._assert_canbe_parent(self)
            self.template._relations += [(self, parent)]
            return parent

        def __rrshift__(self, children):
            """
            Reversed version of `__lshift__()` prepared to locate a list of properties on the left side.

            Parameters
            ----------
            children: [GraphTemplate.Property] | GraphTemplate.Property
                Properties to be children of this property.

            Returns
            -------
            GraphTemplate.Property
                This property.
            """
            self.__lshift__(children)
            return self

    def __init__(self, definitions):
        """
        Construct template with its properties.

        Parameters
        ----------
        definitions: [(str, type, T -> ID, T -> bool)]
            Definitions of template properties.
        """
        self._properties = []
        self._relations = []

        for d in definitions:
            it = iter(d)
            name, kind, ident, ef = (next(it), next(it), next(it, None), next(it, None))

            if isinstance(kind, GraphTemplate):
                prop = GraphTemplate.Property(self, name, kind, None, None)
                _set_template_property(self, prop)
            elif isinstance(kind, GraphTemplate.Property):
                kind.move_template(self, name)
            else:
                _set_template_property(self, GraphTemplate.Property(self, name, kind, ident, ef))

    def __iter__(self):
        """
        Iterate properties in parent-to-child order.

        Returns
        -------
        Iterator[GraphTemplate.Property]
            Property iterator.
        """
        return sort_properties(self._properties)

    def __iadd__(self, another):
        """
        Add another template to this template.

        Parameters
        ----------
        another: GraphTemplate
            Another template.

        Returns
        -------
        GraphTemplate
            This instance.
        """
        for p in another._properties:
            prop = GraphTemplate.Property(self, p.name, p.kind, p.policy, p.entity_filter, origin=p)
            _set_template_property(self, prop)

        for n, p in another._relations:
            getattr(self, n.name) >> getattr(self, p.name)

        return self

    def __add__(self, another):
        """
        Create new template by adding this template and another one.

        Parameters
        ----------
        another: GraphTemplate
            Another template.

        Returns
        -------
        GraphTemplate
            New template.
        """
        template = GraphTemplate([])
        template += self
        template += another

        return template


def _set_template_property(template, prop):
    if hasattr(template, prop.name):
        raise ValueError(f"Property name '{prop.name}' conflicts.'")
    template._properties.append(prop)
    setattr(template, prop.name, prop)


def sort_properties(properties, parent=None):
    names = {p.name for p in properties}

    def walk(p):
        yield p
        for q in p.children:
            for r in walk(q):
                yield r

    for p in filter(lambda p: p.parent is parent, properties):
        for q in walk(p):
            yield q


import sys
from typing import get_type_hints, Generic, TypeVar, List
from inspect import signature, Signature
from .template import GraphTemplate
from .util import chain_serializers, T


if sys.version_info[0:2] >= (3, 8):
    from typing import TypedDict, get_args, get_origin
else:
    class TypedDict(dict):
        pass
    def get_args(tp):
        return tp.__args__
    def get_origin(tp):
        return tp.__origin__


def issubgeneric(t, p):
    """
    Check whether a type is subclass of a generic type.

    Parameters
    ----------
    t: type
        A type to check.
    p: type
        A generic type.

    Returns
    -------
    bool
        Whether the type in subclass of the generic type.
    """
    if type(t) is type:
        return issubclass(t, p)
    else:
        return hasattr(t, '__origin__') and issubclass(get_origin(t), p)


class DocumentedType(Generic[T]):
    """
    Represents a type having documentation as doc-string.
    """
    @staticmethod
    def unpack(dt):
        """
        Unpack a `DocumentedType` into type and documentation.

        Parameters
        ----------
        dt: type(DocumentedType)
            A type of `DocumentedType`.

        Returns
        -------
        type
            Type of the `DocumentedType`.
        str
            Documentation of the `DocumentedType`.
        """
        t = get_args(dt)[0]
        return t, dt.__doc__


def document_type(t, doc):
    """
    Creates a `DocumentedType` for given type and documentation.

    Parameters
    ----------
    t: type
        The type of `DocumentedType`
    doc: string
        The documentation of `DocumentedType`

    Returns
    -------
    Type[DocumentedType]
        Created `DocumentedType`.
    """
    if issubgeneric(t, DocumentedType):
        t = get_args(t)[0]
    class Documented(DocumentedType[T]):
        pass
    Documented[t].__doc__ = doc or ""
    return Documented[t]


class Typeable(Generic[T]):
    """
    An interface for generic type which is resolved into a concrete type by a type parameter.

    Inherit this class and declare static method whose signature is `resolve(me, bound, arg, spec)`.

    >>> class A(Typeable[T]):
    >>>     @staticmethod
    >>>     def resolve(me, bound, arg, spec):
    >>>         ...
    >>>         return some_type
    >>>
    >>> Typeable.resolve(A[T], int, spec)

    Type resolution should start from the invocation of `Typeable.resolve()`.
    It subsequently invokes the static method with arguments below.

    - Typeable type to resolve whose type parameter is replaced with concrete type or another resolved Typeable type.
    - Resolved type for the type parameter.
    - A type passed by the invocation of `Typeable.resolve()`.
    - `GraphSpec` used for schema generation.
    """
    @staticmethod
    def resolve(typeable, arg, spec):
        """
        Resolve a `Typeable` type by a type.

        Parameters
        ----------
        typeable: type[Typeable[T]]
            `Typeable` type having a generic type parameter.
        arg: type
            Type to replace a type variable.
        spec: GraphSpec
            `GraphSpec` used for schema generation.

        Returns
        -------
        type
            Resolved type.
        """
        if get_origin(typeable) is Typeable:
            raise ValueError(f"Typeable should not be used directly. Use inheriting class instead.")

        bound = get_args(typeable)[0]

        if isinstance(bound, TypeVar):
            return Typeable.resolve(typeable[arg], arg, spec)
        elif issubgeneric(bound, Typeable):
            bound = Typeable.resolve(bound, arg, spec)
            return typeable.resolve(typeable, bound, arg, spec)
        else:
            return typeable.resolve(typeable, bound, arg, spec)

    @staticmethod
    def is_resolved(typeable):
        """
        Checks a `Typeable` type is already resolved.

        Parameters
        ----------
        typeable: type[Typeable[T]]
            `Typeable` type having a generic type parameter.

        Returns
        -------
        bool
            Whether the type is already resolved or not.
        """
        bound = get_args(typeable)[0]
        if isinstance(bound, TypeVar):
            return False
        elif issubgeneric(bound, Typeable):
            return Typeable.is_resolved(bound)
        else:
            return True


class DynamicType(Typeable[T]):
    @staticmethod
    def resolve(dynamic, bound, arg, spec):
        return dynamic.fix(bound, arg)

    @classmethod
    def fix(cls, bound, arg):
        return bound


class Shrink(Typeable[T]):
    @staticmethod
    def resolve(shrink, bound, arg, spec):
        if bound == Signature.empty:
            return TypedDict
        if not issubclass(bound, TypedDict):
            raise TypeError(f"Type parameter for Shrink must be resolved to TypedDict but {bound}.")

        class Schema(TypedDict):
            pass
        exc, inc = shrink.select(bound, arg)
        setattr(Schema, '__annotations__', {n:t for n, t in get_type_hints(bound).items() if (not inc or n in inc) and (n not in exc)})

        return Schema

    @classmethod
    def select(cls, bound, arg):
        raise NotImplementedError()


class Extend(Typeable[T]):
    @staticmethod
    def resolve(extend, bound, arg, spec):
        if bound == Signature.empty:
            return TypedDict
        if not issubclass(bound, TypedDict):
            raise TypeError(f"Type parameter for Shrink must be resolved to TypedDict but {bound}.")

        class Schema(extend.schema(bound, arg), bound):
            pass

        return Schema

    @classmethod
    def schema(cls, bound, arg):
        raise NotImplementedError()


def walk_schema(td, with_doc=False):
    """
    Returns a dictionary as a result of walking a schema object from its root.

    Parameters
    ----------
    td: TypedDict
        A schema represented by `TypedDict`.
    with_doc: bool
        Flag to include documentations into result.

    Returns
    -------
    dict
        Dictionary representing the schema.
    """
    if '__annotations__' not in td.__dict__:
        return {}

    result = {}

    def put(k, t, doc):
        if with_doc:
            result[k] = (t, doc)
        else:
            result[k] = t

    def expand(t):
        return (get_args(t)[0], lambda x:[x]) if issubgeneric(t, List) else (t, lambda x:x)

    for k, t in get_type_hints(td).items():
        t, doc = DocumentedType.unpack(t) if issubgeneric(t, DocumentedType) else (t, "")

        t, conv = expand(t)

        if issubclass(t, TypedDict):
            put(k, conv(walk_schema(t, with_doc)), doc)
        else:
            put(k, conv(t), doc)
    
    return result


def templateType(t):
    class Template:
        template = t
    return Template


class GraphSchema:
    """
    This class provides property to get the schema of serialization result of a graph as well as serialization method.
    """
    def __init__(self, spec, template, **serializers):
        self.spec = spec
        self.template = template
        self.serializers = serializers

    def _return_from(self, prop):
        ns = self.serializers[prop.name]

        arg = prop.kind
        arg = templateType(prop.kind) if isinstance(prop.kind, GraphTemplate) else arg

        # Return type of serializers set to NodeSerializer.
        rt = signature(ns.serializer).return_annotation

        # Return type of base serializer obtained from GraphSpec.
        base = chain_serializers(self.spec.find_serializers(arg))

        bt = signature(base).return_annotation if base else Signature.empty
        bt = arg if bt == Signature.empty else bt

        if rt == Signature.empty:
            rt = bt
            bt = arg

        if issubgeneric(bt, Typeable):
            bt = bt if Typeable.is_resolved(bt) else bt[arg]

        if issubgeneric(rt, Typeable):
            if not Typeable.is_resolved(rt):
                rt = rt[bt]
            rt = Typeable.resolve(rt, arg, self.spec)

        return rt

    def schema_of(self, prop):
        """
        Generates structured and documented schema for a template property.

        Parameters
        ----------
        prop: GraphTemplate.Property
            A template property.

        Returns
        -------
        Type[DocumentedType]
            Schema and its documentation.
        """
        rt = self._return_from(prop)

        doc = self.serializers[prop.name]._doc or ""

        if issubclass(rt, dict):
            class Schema(rt):
                pass

            anns = {}

            for c in filter(lambda c: c.name in self.serializers, prop.children):
                ns = self.serializers[c.name]
                cs = self.schema_of(c)

                t, d = DocumentedType.unpack(cs)

                if ns.be_merged:
                    if not issubclass(t, dict):
                        raise ValueError(f"Property '{c.name}' is not configured to be serialized into dict.")
                    anns.update(**{ns.namer(k):t for k, t in get_type_hints(t).items()})
                else:
                    anns[ns.namer(c.name)] = cs if ns.be_singular else document_type(List[t], d)

            setattr(Schema, '__annotations__', anns)

            return document_type(Schema, doc)
        else:
            return document_type(rt, doc)

    @property
    def schema(self):
        """
        Generates `TypedDict` which represents the schema of serialized graph.

        Returns
        -------
        TypedDict
            Representation of the schema of serialized graph.
        """
        class Schema(TypedDict):
            pass

        anns = {}

        def put_root_schema(p):
            nonlocal anns

            ns = self.serializers[p.name]
            dt = self.schema_of(p)

            if ns.be_merged:
                t, d = DocumentedType.unpack(dt)
                anns.update(**{ns.namer(k):t_ for k, t_ in get_type_hints(t).items()})
            elif ns.be_singular:
                anns[ns.namer(p.name)] = dt
            else:
                t, d = DocumentedType.unpack(dt)
                anns[ns.namer(p.name)] = document_type(List[t], d)

        roots = filter(lambda p: p.parent is None and p.name in self.serializers, self.template._properties)

        for p in roots:
            put_root_schema(p)

        setattr(Schema, '__annotations__', anns)

        return Schema

    def serialize(self, graph, **node_params):
        """
        Serialize graph into a dictionary.

        Parameters
        ----------
        graph: GraphView
            A graph.
        node_params: {str: {str: object}}
            Parameters keyed by node names.

        Returns
        -------
        dict
            Serialized representation of the graph.
        """
        return self.spec.to_dict(graph, node_params, **self.serializers)


from .identify import neverPolicy
from .template import sort_properties, GraphTemplate


def new_graph(template, *bases):
    """
    Create a graph from a template.

    Use this function instead of invoking constructor directly.

    Parameters
    ----------
    template: GraphTemplate
        A template of a graph.
    bases: [Graph | GraphView]
        Other graphs whose nodes are appended to created graph.

    Returns
    -------
    Graph
        Created graph.
    """
    graph = Graph(template)

    for b in bases:
        graph += b

    return graph


class Graph:
    """
    The instance of this class contains nodes according to the structure defined by the template.

    All entities should be appended via `append()` or `replace()` and these methods not only store nodes but also construct relationships between them.

    Entity set given on an appending session is arranged according to the structure of graph template.
    Relationships between properties are brought into the graph in the form of parent-child relationships between nodes.

    Additionally, each entity value is compared to entities of existing nodes by identification function set on template property,
    and when some entities of nodes are *identical* to given entity, the graph does not create new node
    and will create relationships between those identical nodes and nodes for child entities.

    For example, following template defines that each entity for property `a` are identified by its value and `a` is the parent of `b`.

    >>> template = GraphSpac().new_template(
    >>>     a = (int, lambda x:x),
    >>>     b = str,
    >>> )
    >>> template.a << template.b

    Invocation of `append()` creates nodes and relationships between them.
    Next code shows that each node of `b` becomes a child of node of `a` created in the same invocation.

    >>> graph = new_graph(template)
    >>> graph.append(a=1, b="a").append(a=2, b="b")
    >>>
    >>> assert [(na(), [nb() for nb in na.b]) for na in graph.view.a] == [(1, ["a"]), (2, ["b"])]

    Meanwhile, when an *identical* entity is given to `a`, node of `a` is not created and existing one is used as the parent of node `b`.

    >>> graph.append(a=1, b="c").append(a=1, b="d")
    >>>
    >>> assert [(na(), [nb() for nb in na.b]) for na in graph.view.a] == [(1, ["a", "c", "d"]), (2, ["b"])]
    """
    def __init__(self, template):
        self.template = template
        self.containers = {p.name:self._to_container(p) for p in template._properties}
        self._view = None

    def _to_container(self, prop):
        if isinstance(prop.kind, GraphTemplate):
            return GraphNodeContainer(prop)
        else:
            return NodeContainer(prop)

    def _container_of(self, prop):
        cands = [c for c in self.containers.values() if c.property.is_compatible(prop)]
        if len(cands) > 1:
            raise ValueError(f"Container can't be determined from property '{prop.name}'.")
        return cands[0] if cands else None

    def __add__(self, another):
        """
        Create new graph by adding this graph and another graph.

        New graph has the same template as this graph'S.
        On the other hand, because this method depends on `__iadd__()`, another graph must not have the same template.

        Parameters
        ----------
        another: Graph | GraphView
            Graph or its view.

        Returns
        -------
        Graph
            Created graph.
        """
        graph = Graph(self.template)

        graph += self
        graph += another

        return graph

    def __iadd__(self, another):
        """
        Append nodes from another graph.

        Templates of this graph and another graph must not be the same.
        Nodes of another graph are traversed from its root and appended to compatible containers each other.

        Parameters
        ----------
        another: Graph | GraphView
            Graph or its view.

        Returns
        -------
        Graph
            This graph.
        """
        another = another if isinstance(another, Graph) else another()

        roots_ = filter(lambda c: c.property.parent is None, another.containers.values())

        def add(n, anc):
            c = self._container_of(n.property)
            if c:
                c.append(n.entity, anc)
            for ch_ in n.children.values():
                for m in ch_.nodes:
                    add(m, anc.copy())

        for c_ in roots_:
            for n_ in c_.nodes:
                add(n_, {})

        return self

    @property
    def view(self):
        """
        Returns an unmodifiable view of this graph.

        Returning object provides intuitive ways to access graph components:

        - Attribute access by the property name returns the view of corresponding node container.
        - Iteration access iterates over views of root containers, that is, containers whose properties have no parent.

        See the documentation of `view` property of `NodeContainer`, `Node`, `Node.Children` for further information.

        Returns
        -------
        GraphView
            The view of this graph.
        """
        if self._view is None:
            graph = self
            class GraphView:
                def __call__(self):
                    """Returns the greph of this view."""
                    return graph
                def __iter__(self):
                    """Iterates views of root containers."""
                    return map(lambda c: (c.name, c.view), filter(lambda c: c.property.parent is None, graph.containers.values()))
                def __getattr__(self, name):
                    """Returns a view of a container of the name."""
                    return graph.containers[name].view
            self._view = GraphView()
        return self._view

    def _append(self, to_replace, entities):
        props = [p for p in self.template if p.name in entities]

        filtered = set()
        for p in props:
            if (p.parent is None) or (p.parent.name not in entities) or (p.parent.name in filtered):
                if p.entity_filter is None or p.entity_filter(entities[p.name]):
                    filtered.add(p.name)

        ancestors = {}
        for k in [p.name for p in props if p.name in filtered]:
            self.containers[k].append(entities[k], ancestors, to_replace)

        return self

    def append(self, **entities):
        """
        Append entity values with associated property names.

        Each key in keyword arguments indicates a property and the paired value is an entity value stored in this graph.

        Parameters
        ----------
        entities: {str: object}
            Dictionary where the key indicates the property name and the value is the entity value.

        Returns
        -------
        Graph
            This graph.
        """
        return self._append(False, entities)

    def replace(self, **entities):
        """
        Behaves like `append()` but replace the entities of identiccal nodes.

        Parameters
        ----------
        entities: {str: object}
            Dictionary where the key indicates the property name and the value is the entity value.

        Returns
        -------
        Graph
            This graph.
        """
        return self._append(True, entities)


class _EmptyNodeView:
    def __init__(self, prop, result):
        self.property = prop
        self.result = result

    def __call__(self, alt=None):
        return self.result

    def __iter__(self):
        return iter([])

    def __getattr__(self, key):
        child = next(filter(lambda c: c.name == key, self.property.children), None)
        if child:
            return _EmptyContainerView(child)
        else:
            raise KeyError(f"Graph property '{base.property.name}' does not have a child property '{key}'.")


class _EmptyContainerView:
    def __init__(self, prop):
        self.property = prop

    def __bool__(self):
        return False

    def __call__(self):
        return []

    def __iter__(self):
        return iter([])

    def __len__(self):
        return 0

    def __getitem__(self, index):
        if isinstance(index, slice):
            return []
        else:
            raise IndexError(f"Index for container '{self.property.name}' is out of range.")

    def __getattr__(self, key):
        child = next(filter(lambda c: c.name == key, self.property.children), None)
        if child:
            return _EmptyContainerView(child)
        else:
            raise KeyError(f"Graph property '{self.property.name}' does not have a child property '{key}'.")


class NodeContainer:
    """
    This class represents a list of nodes for a template property.
    """
    def __init__(self, prop):
        self.nodes = []
        self.keys = {}
        self.property = prop
        self._view = None

    @property
    def name(self):
        """
        Returns a name of the template property.
        """
        return self.property.name

    @property
    def view(self):
        """
        Returns an unmodifiable view of this container.

        Returning object provides intuitive ways to access internal nodes:

        - Iteration access iterates over internal node views.
        - Index access and `len()` works as if it is a list of node views.
        - Attribute access by the property name returns the view of corresponding child node container under the first node.

        See the documentation of `view` property of `Node`, `Node.Children` for further information.

        Returns
        -------
        ContainerView
            The view of this graph.
        """
        if self._view is None:
            container = self
            class ContainerView:
                def __bool__(self):
                    """Returns whether this container is not empty."""
                    return len(container.nodes) != 0
                def __call__(self):
                    """Returns a base container."""
                    return container
                def __len__(self):
                    """Returns the number of nodes."""
                    return len(container.nodes)
                def __iter__(self):
                    """Iterates views of nodes."""
                    return map(lambda n: n.view, container.nodes)
                def __getitem__(self, index):
                    """Returns a view of a node at the index."""
                    if isinstance(index, slice):
                        return [n.view for n in container.nodes[index]]
                    else:
                        return container.nodes[index].view
                def __getattr__(self, key):
                    """Returns a view of the first node or empty container view if it does not exist."""
                    child = next(filter(lambda c: c.name == key, container.property.children), None)
                    if child:
                        return container.nodes[0].children[key].view if len(container.nodes) > 0 else _EmptyContainerView(child)
                    else:
                        raise KeyError(f"Graph property '{container.property.name}' does not have a child property '{key}'.")
            self._view = ContainerView()
        return self._view

    def append(self, entity, ancestors, to_replace=False):
        """
        Add an entity to this container.

        Parameters
        ----------
        entity: object
            An entity to be stored in the node.
        ancestors: {str: [Node]}
            Dictionary which maps property names to nodes appended to ancestor containers.
        to_replace: bool
            If `True`, entities of identical nodes are replaced with given entity.
        """
        def get_nodes(k):
            return [self.nodes[i] for i in self.keys.get(k, [])]

        policy = self.property.policy or neverPolicy()

        key = policy.get_identifier(entity)

        parents, identicals = policy.identify(self.property, [self.nodes[i] for i in self.keys.get(key, [])], ancestors)

        new_nodes = identicals.copy()

        for p in parents:
            index = len(self.nodes)

            node = Node(self.property, entity, key, index)
            self.nodes.append(node)
            if key is not None:
                self.keys.setdefault(key, []).append(index)
            new_nodes.append(node)

            if p is not None:
                p.add_child(node)

        if to_replace:
            for n in identicals:
                n.entity = entity

        ancestors[self.property.name] = new_nodes


class GraphNodeContainer(NodeContainer):
    def append(self, entity, ancestors, to_replace):
        if not isinstance(entity, (dict, Graph)):
            raise ValueError(f"Node of graph only accepts dict or Graph object.")

        policy = self.property.policy or neverPolicy()

        parents, _ = policy.identify(self.property, [], ancestors)

        for p in parents:
            index = len(self.nodes)

            graphs = []

            if p is None or len(p.children[self.name].nodes) == 0:
                g = Graph(self.property.kind)
                node = GraphNode(self.property, g, None, index)
                self.nodes.append(node)

                if p is not None:
                    p.add_child(node)

                graphs.append(g)
            else:
                graphs.extend([n.entity for n in p.children[self.name].nodes])

            for g in graphs:
                if isinstance(entity, dict):
                    g.append(**entity)
                else:
                    g += entity


class Node:
    """
    This class represents a node which contains an entity value.
    """
    class Children:
        def __init__(self, prop):
            self.nodes = []
            self.keys = set()
            self.property = prop
            self._view = None

        @property
        def view(self):
            if self._view is None:
                base = self
                class ChildrenView:
                    def __bool__(self):
                        """Returns whether this container is not empty."""
                        return len(base.nodes) != 0
                    def __call__(self):
                        """Returns children container."""
                        return base
                    def __iter__(self):
                        """Iterates views of child nodes."""
                        return map(lambda n: n.view, base.nodes)
                    def __len__(self):
                        """Returns the number of child nodes."""
                        return len(base.nodes)
                    def __getitem__(self, index):
                        """Returns a view of child node at the index."""
                        if isinstance(index, slice):
                            return [n.view for n in base.nodes[index]]
                        else:
                            return base.nodes[index].view
                    def __getattr__(self, key):
                        """Returns a view of the first node or empty container view if it does not exist."""
                        child = next(filter(lambda c: c.name == key, base.property.children), None)
                        if child:
                            return base.nodes[0].children[key].view if len(base.nodes) > 0 else _EmptyContainerView(child)
                        else:
                            raise KeyError(f"Graph property '{base.property.name}' does not have a child property '{key}'.")
                self._view = ChildrenView()
            return self._view

        def has(self, node):
            return node in self.keys

        def append(self, node):
            if node not in self.keys:
                self.keys.add(node)
                self.nodes.append(node)

    def __init__(self, prop, entity, key, index):
        self.property = prop
        self.entity = entity
        self.key = key
        self.parents = set()
        self.children = {c.name: Node.Children(c) for c in prop.children}
        self._index = index
        self._view = None

    @property
    def name(self):
        return self.property.name

    @property
    def view(self):
        """
        Returns an unmodifiable view of this node.

        Returning object provides intuitive ways to access entity and child nodes:

        - Direct invocation returns the entity value.
        - Iteration access iterates over views of child nodes.
        - Attribute access by the property name returns the view of corresponding child node.

        Returns
        -------
        NodeView
            The view of this node.
        """
        if self._view is None:
            node = self
            class NodeView:
                def __call__(self, alt=None):
                    """Returns an entity of this node."""
                    return node.entity
                def __getattr__(self, name):
                    """Returns a view of child nodes by its name."""
                    return node.children[name].view
                def __iter__(self):
                    """Iterate key-value pairs of child nodes."""
                    return map(lambda nc: (nc[0], nc[1].view), node.children.items())
            self._view = NodeView()
        return self._view

    def add_child(self, child):
        if child.property.template != self.property.template:
            raise ValueError(f"Nodes from difference graph template can't be associated.")
        self.children[child.property.name].append(child)
        child.parents.add(self)
        return self

    def has_child(self, child):
        if child.property.template != self.property.template:
            return False
        elif child.property.name in self.children:
            return child in self.children[child.property.name].keys
        else:
            return False


class GraphNode(Node):
    @property
    def view(self):
        return self.entity.view

    def add_child(self, child):
        raise TypeError(f"GraphNode does not have child.")

    def has_child(self, child):
        return False


class Configurable:
    """
    Interface for classes of configuration objects.
    """
    def clone(self):
        raise NotImplementedError()

    def replace(self, another):
        raise NotImplementedError()


def key_to_index(values, ordered_keys):
    """
    Generates a dictionary whose values are same as given dictionary but each key is an index of original key in the ordered key list.

    Parameters
    ----------
    values: {(str|int): object}
        A dictionary.
    ordered_keys: [str]
        Ordered key list. If some original keys are not contained in this list, `ValueError` raises.
    """
    def index(k):
        if isinstance(k, int):
            return k
        else:
            return ordered_keys.index(k)

    return {index(k):v for k, v in values.items()}


from datetime import datetime, date, time, timedelta
from ..config import default_config


def config(cfg=default_config().derive()):
    return cfg


def truncate(db, *models):
    """
    Truncate tables in order.

    Parameters
    ----------
    db: Connection
        DB connection.
    tables: [model]
        Models of tables.
    """
    if len(models) == 0:
        raise ValueError(f"No tables are specified. Did you forget to pass DB connection at the first argument?")

    for m in models:
        m.truncate(db)


class Matcher:
    def __init__(self):
        self.invert = False

    def __invert__(self):
        self.invert = True
        return self

    def __and__(self, another):
        return CompositeMatcher(self, another, True)

    def __or__(self, another):
        return CompositeMatcher(self, another, False)

    def match(self, actual):
        return self._match(actual) ^ self.invert

    def _match(self, actual):
        raise NotImplementedError()


class CompositeMatcher(Matcher):
    def __init__(self, m1, m2, and_=True):
        super().__init__()
        self.m1 = m1
        self.m2 = m2
        self.and_ = and_

    def _match(self, actual):
        if self.and_:
            return self.m1.match(actual) and self.m2.match(actual)
        else:
            return self.m1.match(actual) or self.m2.match(actual)


class Near(Matcher):
    def __init__(self, expected, negative=None, positive=None, **kwargs):
        super().__init__()
        self.expected = expected
        self.negative = negative
        self.positive = positive
        self.kwargs = kwargs

    def _margin(self, t):
        neg = self.negative or 0
        pos = self.positive or 0

        if issubclass(t, (int, float)):
            return (self.expected + neg, self.expected + pos)
        elif issubclass(t, (datetime, date)):
            delta_args = {k:v for k, v in self.kwargs.items() if k in {
                'weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds', 'microseconds',
            }}
            delta = timedelta(**delta_args)
            return (self.expected + neg * delta, self.expected + pos * delta)
        else:
            raise ValueError(f"Margin for {t} is not supported.")

    def _match(self, actual):
        if self.negative is None and self.positive is None:
            return actual == self.expected
        else:
            low, high = self._margin(type(actual))
            return low <= actual and actual <= high


def near(expected, negative=None, positive=None, **kwargs):
    """
    Create a matcher to check actual value is conatined in a range.

    Parameters
    ----------
    expected: object
        Expected value.
    negative: object
        Margin of negative direction.
    positive: object
        Margin of positive direction.
    kwargs: {str: object}
        Keyword arguments to create marginal values.

    Returns
    -------
    Near
        Matcher.
    """
    if isinstance(expected, datetime):
        if all(k not in kwargs for k in ('weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds', 'microseconds')):
            kwargs.update(**config().timedelta_unit)
    return Near(expected, negative, positive, **kwargs)


class let(Matcher):
    def __init__(self, pred):
        """
        Create a matcher which applies the predicate function to actual value and checks its returning value is `True`.

        Parameters
        ----------
        pred: object -> bool
            A predicate function.

        Returns
        -------
        let
            Matcher.
        """
        super().__init__()
        self.pred = pred

    def _match(self, actual):
        return self.pred(actual)


class one_of(Matcher):
    def __init__(self, *candidates):
        """
        Create a matcher which checks whether the actual value matches one of candidate values.

        Parameters
        ----------
        candidates: [object]
            Candidate values.

        Returns
        -------
        one_of
            Matcher.
        """
        super().__init__()
        self.candidates = candidates

    def _match(self, actual):
        for c in self.candidates:
            if isinstance(c, Matcher):
                if c.match(actual):
                    return True
            elif c == actual:
                return True
        return False


from collections import OrderedDict


class Column:
    """
    This class represents a schema of a column.

    Attributes
    ----------
    name: str
        Column name.
    ptype: type
        Data type in python.
    type_info: str
        Type informations obtained from DB.
    pk: bool
        Is this column a PK?
    fk: object
        An informative object if this column is a foreign key, otherwise `None`.
    incremental: object
        If this column is auto-incremental, this object contains the information of the feature, otherwise, None.
    comment: str
        Comment of the column.
    """
    def __init__(self, name, ptype, type_info, pk, fk, incremental, comment=""):
        self.name = name
        self.ptype = ptype
        self.type_info = type_info
        self.pk = pk
        self.fk = fk
        self.incremental = incremental
        self.comment = comment


class Table:
    """
    This class represents a schema of a table.

    Attributes
    ----------
    name: str
        Table name.
    columns: [Column]
        Columns in the table.
    comment: str
        Comment of the table.
    """
    def __init__(self, name, columns, comment=""):
        self.name = name
        self.columns = columns
        self.comment = ""


def define_model(table_, mixins=[]):
    """
    Create a model type representing the table.

    Types in `mixins` argument are base types which created type inherits.
    When the same attribute is defined in multiple mixin types, the former overwrites the latter.

    Model type always has following attributes.

    - name: str
        Name of the table.
    - table: Table
        Table schema.
    - columns: [Column]
        The list of column schemas.
    - column: object
        An object exposing schemas of columns via attributes of their names.

    Model instances are created by passing column name and its value as keyword arguments to the constructor.
    They are plain python objects representing a DB record by holding the subset of its columns as attributes.

    >>> # CREATE TABLE t1 (col1 int, col2 text, col3 text);
    >>> table = define_model("t1")
    >>> model = table(col1=1, col2="a")

    Attributes are also assignable by normal setter. In any case, attribute name must be a one of column names, otherwise `TypeError` raises.

    >>> model.col3 = "b"

    Model instance supports iteration which yields pairs of assigned column schema and its value.

    >>> for c, v in model:
    >>>     print(f"{c.name} = {v}")
    >>>
    col1 = 1
    col2 = a
    col3 = b

    Parameters
    ----------
    table_: Table
        Schema of table.
    mixins: [type]
        Types the created model type inherits.

    Returns
    -------
    type
        Created model type.
    """
    column_names = {c.name for c in table_.columns}

    class Meta(type):
        name = table_.name
        table = table_
        columns = table_.columns

        @classmethod
        def shrink(cls, excludes, includes=None):
            """
            Creates new model type containing subset of columns.

            Parameters
            ----------
            excludes: [str]
                Column names to exclude.
            includes: [str]
                Column names to include.

            Returns
            -------
            type
                Created model type.
            """
            cols = [c for c in cls.columns if (not includes or c.name in includes) and c.name not in excludes]
            return define_model(Table(cls.name, cols, cls.table.comment), mixins)

    class Columns:
        def __init__(self):
            for c in table_.columns:
                setattr(self, c.name, c)

    setattr(Meta, "column", Columns())

    class Base(metaclass=Meta):
        pass

    class Model(type("ModelBase", tuple([Base] + mixins), {})):
        def __init__(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)

        def __repr__(self):
            cls = type(self)
            return f"{cls.name}({', '.join([f'{c.name}={repr(getattr(self, c.name))}' for c in cls.columns if hasattr(self, c.name)])})"

        def __str__(self):
            cls = type(self)
            return f"{cls.name}({', '.join([f'{c.name}={str(getattr(self, c.name))}' for c in cls.columns if hasattr(self, c.name)])})"

        def __iter__(self):
            cls = type(self)
            return map(lambda c: (c, getattr(self, c.name)), filter(lambda c: hasattr(self, c.name), cls.columns))

        def __setattr__(self, key, value):
            if key not in column_names:
                raise TypeError(f"{key} is not a column of {type(self).name}")
            object.__setattr__(self, key, value)

        def __getitem__(self, key):
            return getattr(self, key)

        def __contains__(self, key):
            return hasattr(self, key)

        def __eq__(self, other):
            cls = type(self)
            if cls != type(other):
                return False
            for k in column_names:
                if hasattr(self, k) ^ hasattr(other, k):
                    return False
                if getattr(self, k, None) != getattr(other, k, None):
                    return False
            return True

    return Model


def parse_pks(model, pks):
    """
    Generates a pair of PK columns names and their values from polymorphic input.

    Parameters
    ----------
    model: type
        Model class.
    pks: object | {str: object}
        A dictionary of PK column name and their values or an object of single PK column.

    Returns
    -------
    [str]
        Names of PK columns.
    [object]
        Values of PK columns.
    """
    if isinstance(pks, dict):
        check_columns(model, pks, lambda c: c.pk, True)
        return list(pks.keys()), list(pks.values())
    else:
        cols = [c.name for c in model.columns if c.pk]
        if len(cols) != 1:
            raise ValueError(f"The number of primary key columns in {model.name} is not 1.")
        return ([cols[0]], [pks])


def check_columns(model, col_map, condition=lambda c: True, requires_all=False):
    """
    Checks keys of given `dict` match columns selected by a condition from a model.

    Parameters
    ----------
    model: type
        Model class.
    col_map: {str: object}
        Dictionary whose keys are column names.
    condition: Column -> boolean
        A Function which selects columns from the model.
    requires_all: boolean
        If `True`, `ValueError` raises when the dictionary does not contain keys of all selected columns.
    """
    names = set([c.name for c in model.columns if condition(c)])
    targets = set(col_map.keys())
    if not names >= targets:
        raise ValueError(f"Columns {targets - names} are not columns of '{model.name}'.")
    if requires_all and not names == targets:
        raise ValueError(f"Required columns {names - targets} in '{model.name}' are not found.")


def model_values(model, values, excludes_pk=False):
    """
    Generates a dictionary whose items are pairs of column name and column value.

    Parameters
    ----------
    model: type
        Model class.
    values: dict | Model
        Dictionary from column name to column value or a model instance.
    excludes_pk: boolean
        If `True`, item of PK column is not contained in returned dictionary.

    Returns
    -------
    dict
        A dictionary from column name to column value.
    """
    if isinstance(values, (dict, OrderedDict)):
        includes = {c.name for c in model.columns if (not excludes_pk) or (not c.pk)}
        return {k:v for k, v in values.items() if k in includes}
    elif isinstance(values, model):
        return {cv[0].name:cv[1] for cv in values if (not excludes_pk) or (not cv[0].pk)}
    else:
        raise TypeError(f"Values to insert or update must be a dictionary or model.")


import logging
from functools import wraps
from .model_graph import ConfigurableSpec
from .util import Configurable


class PyracmonConfiguration:
    """
    Configurations for various modules.

    Attributes
    ----------
    name: str
        Name of this configuration. This value has no effect on any behavior of modules.
    logger: str | logging.Logger
        Logger or the name of logger used for internal logs such as query logging.
    log_level: int
        Logging level of internal logs.
    sql_log_length: int
        Maximum length for query log. Queries longer than this value are output with being trimmed.
    parameter_log: bool
        Flag to log query parameters also.
    paramstyle: str
        Parameter style defined in DB-API 2.0. This value overwrites the style obtained via DB module.
    type_mapping: (str, **) -> type
        Function estimating python type from type name in database and optional DBMS dependent keys.
    graph_spec: ConfigurableSpec
        Graph specification used as default.
    fixture_mapping: (Table, Column, int) -> object
        Function generating fixture value for a column and an index.
    fixture_tz_aware: bool
        Flag to make fixture datetime being aware of timezone.
    timedelta_unit: dict
        Default keyword arguments to pass `timedelta()` to compare actual `date`/`datetime` with expected one in `near` matcher.
    """
    def __init__(
        self,
        name = None,
        logger = None,
        log_level = None,
        sql_log_length = None,
        parameter_log = None,
        paramstyle = None,
        type_mapping = None,
        graph_spec = None,
        fixture_mapping = None,
        fixture_tz_aware = None,
        timedelta_unit = None,
    ):
        self.name = name
        self.logger = logger
        self.log_level = log_level
        self.sql_log_length = sql_log_length
        self.parameter_log = parameter_log
        self.paramstyle = paramstyle
        self.type_mapping = type_mapping
        self.graph_spec = graph_spec or ConfigurableSpec.create()
        self.fixture_mapping = fixture_mapping
        self.fixture_tz_aware = fixture_tz_aware
        self.timedelta_unit = timedelta_unit

    def derive(self, **kwargs):
        def attr(k):
            v = getattr(self, k)
            return v.clone() if isinstance(v, Configurable) else None
        attrs = {k:attr(k) for k in vars(self) if attr(k) is not None}
        attrs.update(kwargs)
        return DerivingConfiguration(self, **attrs)


class DerivingConfiguration(PyracmonConfiguration):
    def __init__(self, base, **kwargs):
        super().__init__(**kwargs)
        self.base = base

    def __getattribute__(self, key):
        value = object.__getattribute__(self, key)
        return value if value is not None else getattr(self.base, key)

    def set(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self.base, k):
                setattr(self, k, v)
            else:
                raise KeyError(f"Unknown configuration key: {k}")
        return self


def default_config(config=PyracmonConfiguration(
    name = "default",
    logger = None,
    log_level = logging.DEBUG,
    sql_log_length = 4096,
    parameter_log = False,
    paramstyle = None,
    type_mapping = None,
    graph_spec = None,
    fixture_mapping = None,
    fixture_tz_aware = True,
    timedelta_unit = dict(seconds=1),
)):
    return config


def pyracmon(**kwargs):
    """
    Generates an object which starts `with` context, where global configuration can be changed.

    An object obtained by `as` keyword in `with` clause is an instance of `PyracmonConfiguration`,
    and changes to it done in the `with` context are copied into global configuration at the end of the context if no exception raises.

    >>> with pyracmon() as cfg:
    >>>     cfg.name = "my_config"
    >>>     ...
    """
    class Configurable:
        def __init__(self):
            self.config = default_config().derive()

        def __enter__(self):
            return self.config

        def __exit__(self, exc_type, exc_value, traceback):
            if not exc_value:
                target = default_config()

                for k in vars(target):
                    v = getattr(self.config, k)
                    if isinstance(v, Configurable):
                        getattr(target, k).replace(v)
                    elif v is not None:
                        setattr(target, k, v)
            return False

    return Configurable()




import logging
from .config import default_config


class ConnectionContext:
    """
    This class represents a context where DB operation is done on the configuration it has.

    By default, each instance has the local copy of global configuration which can be changed via `configure()` 
    """
    def __init__(self, identifier=None, **configurations):
        self.identifier = identifier
        self.config = default_config().derive(**configurations)

    def _message(self, message):
        return f"({self.identifier}) {message}" if self.identifier else message

    def configure(self, **configurations):
        """
        Change the configuration of this context.

        Parameters
        ----------
        configurations: {str: object}
            Pairs of configuration name and its value.

        Returns
        -------
        ConnectionContext
            This instance.
        """
        self.config.set(**configurations)
        return self

    def execute(self, cursor, sql, params):
        """
        Executes a query on a cursor.

        Parameters
        ----------
        cursor: Cursor
            Cursor object.
        sql: str
            Query string.
        params: [object]
            Query parameters.

        Returns
        -------
        Cursor
            Given cursor object. Internal state may be changed by the execution of the query.
        """
        logger = _logger(self.config)

        if logger:
            sql_log = sql if len(sql) <= self.config.sql_log_length else f"{sql[0:self.config.sql_log_length]}..."

            logger.log(self.config.log_level, self._message(sql_log))

            if self.config.parameter_log:
                logger.log(self.config.log_level, self._message(f"Parameters: {params}"))

        cursor.execute(sql, params)

        return cursor


def _logger(config):
    if isinstance(config.logger, logging.Logger):
        return config.logger
    elif isinstance(config.logger, str):
        return logging.getLogger(config.logger)
    else:
        return None